#!usr/bin/env python
import os
import glob
modules = glob.glob(os.path.dirname(__file__)+"/*.py")
__all__ = [ os.path.basename(f)[:-3] for f in modules]

_package_dir = os.path.dirname(os.path.realpath(__file__))
__version__ = '0.4.0'

def load_example_data():
    import pandas as pd
    _examples_dir = os.path.join(_package_dir, 'docs/examples')
    data_fpath = os.path.join(_examples_dir, 'reactive_example_idx.csv')
    return pd.read_csv(data_fpath).copy()

def load_dpm_animation():
    import io, base64
    from IPython.display import HTML
    _examples_dir = os.path.join(_package_dir, 'docs/examples')
    mov_fpath = os.path.join(_examples_dir, 'anim.mp4')
    video = io.open(mov_fpath, 'r+b').read()
    encoded = base64.b64encode(video)
    data='''<video width="50%" alt="test" loop=1 controls> <source src="data:video/mp4; base64,{0}" type="video/mp4" /> </video>'''.format(encoded.decode('ascii'))
    return HTML(data=data)

def style_notebook():
    from IPython.core.display import HTML
    g3link = "(https://www.github.com/dunovank/jupyter-themes)"
    #print("Notebook Theme: Grade3\n{}".format(g3link))
    _styles_dir = os.path.join(_package_dir, 'styles')
    style = os.path.join(_styles_dir, 'custom.css')
    csscontent = open(style, "r").read()
    return HTML(csscontent)



#!/usr/local/bin/env python
# Authors:
#   Kyle Dunovan (ked64@pitt.edu) &
#   Jeremy Huang (jeremyhuang@cmu.edu)
from __future__ import division
import numpy as np
import numba as nb
from scipy.stats import norm
from numba.decorators import jit
from numba import float64, int64, vectorize, boolean
from numpy.random import random_sample as randsample


def sample_posterior_belief(data, m0=300, k0=1, s2=1, v0=1, nsamples=800):

    ntrials = data.shape[0]
    muPosterior = np.zeros((ntrials, nsamples))
    varPosterior = np.zeros((ntrials, nsamples))

    for i in range(1, ntrials+1):
        mu_t, var_t = draw_mus_and_sigmas(data[:i], m0, k0, s2, v0, nsamples)
        muPosterior[i-1, :] = mu_t
        varPosterior[i-1, :] = var_t

    return muPosterior, varPosterior


def rolling_variance(old_avg, new, old, N):
    newavg = oldavg + (new - old)/N
    average = newavg
    variance += (new-old)*(new-newavg+old-oldavg)/(N-1)
    stddev = sqrt(variance)


def get_norm(mu, sd, sample):
    return norm(mu, sd).logpdf(sample)


@vectorize([float64(boolean, float64, float64)])
def ufunc_where(condition, x, y):
    if condition:
        return x
    else:
        return y

@vectorize([int64(float64, float64)], nopython=True)
def get_onset_index(onset, dt):
    onsets = int(onset / dt)
    return onsets


@jit(nopython=True)
def slice_theta_array(theta_array, index_arrays, n_vals):
    i, ix = 0, 0
    param_array = np.empty(index_arrays.T.shape)
    for nlvls in n_vals:
        arr_ix = index_arrays[i]
        param_array[:, i] = theta_array[ix:ix+nlvls][arr_ix]
        i += 1
        ix += nlvls
    return param_array.T



@jit(nb.typeof((1, 1))(float64[:], float64[:], float64, float64, float64, float64), nopython=True)
def sim_ddm_trace(rProb, trace, vProb, bound, gbase, dx):
    evidence = gbase
    trace[0] = evidence
    timebound = trace.size
    for ix in range(1, timebound):
        if rProb[ix] < vProb:
            evidence += dx
        else:
            evidence -= dx
        trace[ix] = evidence
        if evidence >= bound:
            return ix, 1
        elif evidence <= 0:
            return ix, 0
    return -1, -1


@jit((float64[:,:,:], float64[:,:,:], float64[:,:], float64[:,:], float64[:], float64[:], float64[:], int64[:], float64[:], float64), nopython=True)
def sim_many_ddm_traces(rProb, dvg, rts, choices, vProb, bound, gbase, gOnset, dx, dt):
    ncond, ntrials, ntime = rProb.shape
    for i in range(ncond):
        tr = gOnset[i]
        dvg[i,:,:tr] = gbase[i]
        for j in range(ntrials):
            ix, choice = sim_ddm_trace(rProb[i,j,tr:], dvg[i,j,tr:], vProb[i], bound[i], gbase[i], dx[i])
            if ix<0:
                rt_ix = ntime + 1
                rts[i,j] = 1000.
            else:
                rt_ix = tr + ix
                rts[i,j] = rt_ix * dt
            choices[i, j] = choice



@jit(int64(float64[:], float64[:], float64[:], float64, float64, float64, float64), nopython=True)
def sim_dpm_trace_upper(rProb, trace, xtb, vProb, bound, gbase, dx):
    evidence = gbase
    trace[0] = evidence
    timebound = trace.size
    for ix in range(1, timebound):
        if rProb[ix] < vProb:
            evidence += dx
        else:
            evidence -= dx
        weightedEvidence = evidence * xtb[ix]
        trace[ix] = weightedEvidence
        if weightedEvidence >= bound:
            return ix
    return -1


@jit(float64(float64[:], float64, float64, int64, float64, float64), nopython=True)
def sim_dpm_trace_lower(rProbSS, ssbase, vsProb, onset, dx, dt):
    ix = onset
    evidence = ssbase
    timebound = rProbSS.size
    while evidence>0 and ix<timebound:
        ix += 1
        if rProbSS[ix] < vsProb:
            evidence += dx
            continue
        evidence -= dx
    return ix * dt


@jit((float64[:,:,:], float64[:,:,:,:], float64[:,:,:], float64[:,:], float64[:,:,:], float64[:,:], float64[:], float64[:], float64[:], float64[:], int64[:], int64[:,:], float64[:], float64[:], float64), nopython=True)
def sim_many_dpm(rProb, rProbSS, dvg, rts, ssrts, xtb, drift, ssdrift, bound, gbase, gOnset, ssOnset, dx, si, dt):
    ncond, ntrials, ntime = rProb.shape
    ncond, nssd, nss_per, ntime = rProbSS.shape
    vProb = 0.5 * (1 + (drift * np.sqrt(dt))/si)
    vsProb = 0.5 * (1 + (ssdrift * np.sqrt(dt))/si)
    for i in range(ncond):
        tr = gOnset[i]
        ssOn = ssOnset[i]
        dvg[i,:,:tr] = gbase[i]
        for j in range(ntrials):
            ix = sim_dpm_trace_upper(rProb[i,j,tr:], dvg[i,j,tr:], xtb[i], vProb[i], bound[i], gbase[i], dx[i])
            if ix<0:
                rt_ix = ntime + 1
                rts[i,j] = rt_ix * dt # 1000.
            else:
                rt_ix = tr + ix
                rts[i,j] = rt_ix * dt
            # Simulate Stop Process
            if j<nss_per:
                ssbase = dvg[i,j][ssOn]
                for k in range(nssd):
                    if rt_ix < ssOn[k] or ix<0:
                        ssrts[i,k,j] = ntime * dt # 1000.
                        continue
                    ssrts[i,k,j] = sim_dpm_trace_lower(rProbSS[i,k,j], ssbase[k], vsProb[i], ssOn[k], dx[i], dt)



@jit(float64(float64[:], float64[:], float64, float64, int64, float64, float64), nopython=True)
def sim_dpm_trace_lower_trace(rProbSS, dvs, ssbase, vsProb, onset, dx, dt):
    ix = onset
    evidence = ssbase
    timebound = rProbSS.size
    while evidence>0 and ix<timebound:
        if rProbSS[ix] < vsProb:
            evidence += dx
        else:
            evidence -= dx
        dvs[ix] = evidence
        ix += 1
    return ix * dt


@jit((float64[:,:,:], float64[:,:,:,:], float64[:,:,:], float64[:,:,:,:], float64[:,:], float64[:,:,:], float64[:,:], float64[:], float64[:], float64[:], float64[:], int64[:], int64[:,:], float64[:], float64[:], float64), nopython=True)
def sim_many_dpm_traces(rProb, rProbSS, dvg, dvs, rts, ssrts, xtb, drift, ssdrift, bound, gbase, gOnset, ssOnset, dx, si, dt):
    ncond, ntrials, ntime = rProb.shape
    ncond, nssd, nss_per, ntime = rProbSS.shape
    vProb = 0.5 * (1 + (drift * np.sqrt(dt))/si)
    vsProb = 0.5 * (1 + (ssdrift * np.sqrt(dt))/si)
    for i in range(ncond):
        tr = gOnset[i]
        ssOn = ssOnset[i]
        dvg[i,:,:tr] = gbase[i]
        for j in range(ntrials):
            ix = sim_dpm_trace_upper(rProb[i,j,tr:], dvg[i,j,tr:], xtb[i], vProb[i], bound[i], gbase[i], dx[i])
            if ix<0:
                rt_ix = ntime + 1
                rts[i,j] = 1000.
            else:
                rt_ix = tr + ix
                rts[i,j] = rt_ix * dt
            # Simulate Stop Process
            if j<nss_per:
                ssbase = dvg[i,j][ssOn]
                for k in range(nssd):
                    if rt_ix < ssOn[k] or ix<0:
                        ssrts[i,k,j] = 1000.
                        continue
                    ssrts[i,k,j] = sim_dpm_trace_lower_trace(rProbSS[i,k,j], dvs[i,k,j], ssbase[k], vsProb[i], ssOn[k], dx[i], dt)




@jit(nb.typeof((1.0, 1.0))(float64[:], float64[:], float64, float64, int64, float64[:], float64, int64, float64, float64, float64, float64), nopython=True)
def sim_dpm_go_stop(rProb, xtb, vProb, bound, onsetIX, rProbSS, vsProb, ssdIX, dx, dt, evidence, ssbound):
    ssEvidence = 0
    ssStarted = 0
    weightedEvidence = 0
    timebound = rProb.shape[0]
    start = onsetIX
    if ssdIX <= onsetIX:
        ssStarted = 1
        start = ssdIX
    for ix in range(start, timebound):
        if rProb[ix] < vProb:
            evidence += dx
        else:
            evidence -= dx
        weightedEvidence = evidence * xtb[ix]
        if weightedEvidence >= bound:
            return 1., ix*dt
        if ix == ssdIX:
            ssStarted = 1
            ssEvidence = weightedEvidence
        if ssStarted:
            if rProbSS[ix] < vsProb:
                ssEvidence += dx
            else:
                ssEvidence -= dx
            if ssEvidence <= ssbound:
                return 0., ix*dt
    return 0., ix*dt


@jit(nb.typeof((1.0, 1.0))(float64[:], float64[:], float64, float64, int64, float64, float64, float64), nopython=True)
def sim_dpm_go(rProb, xtb, vProb, bound, onsetIX, dx, dt, evidence):
    timebound = rProb.shape[0]
    for ix in range(onsetIX, timebound):
        if rProb[ix] < vProb:
            evidence += dx
        else:
            evidence -= dx
        weightedEvidence = evidence * xtb[ix]
        if weightedEvidence >= bound:
            return 1., ix*dt
    return 0., ix*dt



@jit((float64[:,:], float64[:,:], float64[:,:], float64[:], int64[:], float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, int64), nopython=True)
def sim_dpm_learning(results, rProb, rProbSS, xtb, idxArray, drift, ssdrift, bound, onset, B, C, R, dx, dt, tb, maxTrials):

    onsetIX = np.int(onset)
    TargetRT = .52
    goStartTrial = 0.
    ssbound = 0.
    si = .1
    tb = tb + .02

    vsProb = 0.5 * (1 + (ssdrift * np.sqrt(dt))/si)

    for idx in idxArray:
        idxResults = results[results[:, 0]==idx]
        ntrials = idxResults.shape[0]
        vTrial = drift
        aTrial = bound
        errT = 10000

        for t in range(ntrials):
            ttype, ssOnset = idxResults[t, 1:3]
            sensitivity = np.exp2(-t*R) #(t/10.)**-R
            B_t = B * sensitivity
            C_t = C * sensitivity

            vProbTrial = 0.5 * (1 + (vTrial * np.sqrt(dt))/si)

            if ttype==0.:
                ssdIX = np.int(ssOnset)
                response, rt = sim_dpm_go_stop(rProb[t], xtb, vProbTrial, aTrial, onsetIX, rProbSS[t], vsProb, ssdIX, dx, dt, goStartTrial, ssbound)

                if response:
                    correct = 0.; score = -1.; errT = 0.
                    #vTrial = vTrial * np.exp((rt-tb) * B_t)
                    #vTrial = vTrial + B_t * ((aTrial/TargetRT) - (aTrial/rt))
                    vTrial = vTrial + B_t * ((aTrial/tb) - (aTrial/rt))
                else:
                    correct = 1.; score = 0.; errT += 1.
                aTrial = bound + C_t * np.exp(-errT)

            else:
                errT += 1
                response, rt = sim_dpm_go(rProb[t], xtb, vProbTrial, aTrial, onsetIX, dx, dt, goStartTrial)
                if response:
                    correct = 1.
                    score = np.exp(-abs(TargetRT-rt)*10)
                    #vTrial = vTrial * np.exp((rt-TargetRT) * B)
                    vTrial = vTrial + B * ((aTrial/TargetRT) - (aTrial/rt))
                else:
                    correct = 0.
                    score = -1.

            if vTrial > 2.0:
                vTrial = 1.8
            elif vTrial < .25:
                vTrial = .28

            idxResults[t] = np.array([idx, ttype, ssOnset, response, correct, rt, sensitivity, vTrial, aTrial])

        results[results[:, 0]==idx] = idxResults


#
# @jit((float64[:,:], float64[:,:], float64[:,:], float64[:], int64[:], float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, int64), nopython=True)
# def sim_dpm_learning(results, rProb, rProbSS, xtb, idxArray, vProb, vsProb, bound, onset, B, C, R, dx, dt, tb, maxTrials):
#
#     # rProb = randsample((maxTrials, 680))
#     # rProbSS = randsample((maxTrials, 680))
#     onsetIX = np.int(onset)
#     TargetRT = .52
#     goStartTrial = 0.
#     ssbound = 0.
#     # nblocks = rtMatrix.shape[1]
#     # resultsTemp = results
#     # if makeTemp:
#     # resultsTemp = np.zeros_like(results)
#     # resultsTemp[:, np.array([0, 1, 2])] = results[:, np.array([0, 1, 2])]
#
#     # Taim = lambda tr, a, v: tr + (a / v)
#     # zBrake = lambda v, muSSD, tr: v * (muSSD - tr)
#     # SSRT = lambda z, muSSD, ssv: muSSD + (z / abs(ssv))
#
#     for idx in idxArray:
#         idxResults = results[results[:, 0]==idx]
#         ntrials = idxResults.shape[0]
#         vTrial = vProb
#         aTrial = bound
#         errT = 10000
#
#         sigTrial = idxResults[:, 2].std()
#         tau = sigTrial
#         alphaTrial = sigTrial / (sigTrial + tau)
#
#         for t in range(ntrials):
#             ttype, ssOnset = idxResults[t, 1:3]
#             sensitivity = np.exp2(-t*R) #(t/10.)**-R
#             B_t = B * sensitivity
#             C_t = C * sensitivity
#             if ttype==0.:
#                 ssdIX = np.int(ssOnset)
#                 response, rt = sim_dpm_go_stop(rProb[t], xtb, vTrial, aTrial, onsetIX, rProbSS[t], vsProb, ssdIX, dx, dt, goStartTrial, ssbound)
#
#                 if response:
#                     correct = 0.; score = -1.; errT = 0.
#                     #vTrial = vTrial * np.exp((rt-tb) * B_t)
#                     vTrial = vTrial + C_t * ((aTrial/tb) - (aTrial/rt))
#                 else:
#                     correct = 1.; score = 0.; errT += 1.
#
#                 aTrial = bound + C_t * np.exp(-errT)
#                 sigTrial = sigTrial - (alphaTrial * sigTrial)
#                 alphaTrial = sigTrial / (sigTrial + tau)
#
#                 # aTrial =
#             else:
#                 errT += 1
#                 response, rt = sim_dpm_go(rProb[t], xtb, vTrial, aTrial, onsetIX, dx, dt, goStartTrial)
#                 if response:
#                     correct = 1.
#                     score = np.exp(-abs(TargetRT-rt)*10)
#                     # vTrial = vTrial * np.exp((rt-TargetRT) * B)
#
#                     #############################
#                     ### NEW DRIFT UPDATE RULE ###
#                     #############################
#                     vTrial = vTrial + B_t * ((aTrial/TargetRT) - (aTrial/rt))
#                     #vTrial = vTrial + alphaTrial * ((aTrial/rt) - (aTrial/TargetRT))
#                 else:
#                     correct = 0.
#                     score = -1.
#             if vTrial > 1.0:
#                 vTrial = 1.
#             elif vTrial < .05:
#                 vTrial = .05
#             idxResults[t] = np.array([idx, ttype, ssOnset, response, correct, rt, alphaTrial, vTrial, aTrial, sigTrial, score])
#         results[results[:, 0]==idx] = idxResults

    # return results
        # rtVals = idxResults[idxResults[:,3]==1.][:, 5]
        # saccVals = idxResults[idxResults[:,1]==0.][:, 4]
        # ixRT = 0; ixAcc = 0
        # wRT = np.int(np.floor(rtVals.size / nblocks))
        # wAcc = np.int(np.floor(saccVals.size / nblocks))
        # for i in range(nblocks):
        #     rtMatrix[idxCount, i] = np.nanmean(rtVals[ixRT:ixRT+wRT])
        #     saccMatrix[idxCount, i] = np.nanmean(saccVals[ixAcc:ixAcc+wAcc])
        #     ixRT += wRT
        #     ixAcc += wAcc
        # idxCount += 1


# @jit(float64[:,:](float64[:,:], float64[:,:], float64[:,:], float64[:], int64[:], float64, float64, float64, float64, float64, float64, float64, float64, float64, float64, int64, int64), nopython=True)
# def sim_dpm_learning_nruns(nresults, rProb, rProbSS, xtb, idxArray, vProb, vsProb, bound, onset, B, C, R, dx, dt, tb, ntrials, n):
#     nidx = idxArray.size
#     nrowsTotal = nidx * ntrials
#     for i in range(n):
#         start = nrowsTotal*i
#         end = nrowsTotal*(i+1)
#         res = nresults[start:end, :]
#         nresults[start:end, :] = sim_dpm_learning(res, rProb, rProbSS, xtb, idxArray, vProb, vsProb, bound, onset, B, C, R, dx, dt, tb, ntrials)
#     return nresults


@jit((float64[:,:,:], float64[:,:], float64[:,:], float64[:], float64[:], int64[:], float64[:], float64), nopython=True)
def sim_many_single(rProb, rts, xtb, vProb, bound, gOnset, dx, dt):
    ncond, ntrials, ntime = rProb.shape
    tb = ntime * dt
    for i in range(ncond):
        tr = gOnset[i]
        threshold = bound[i]
        timebound = ntime-tr
        for j in range(ntrials):
            ix = tr
            urg_ix = 0
            evidence = 0
            weightedEvidence = 0
            while weightedEvidence<threshold and ix<timebound:
                ix += 1
                urg_ix += 1
                if rProb[i, j, ix] < vProb[i]:
                    evidence += dx[i]
                    continue
                evidence -= dx[i]
                weightedEvidence = evidence * xtb[i, urg_ix]
            rts[i,j] = ix * dt


def sim_multi_params(rProb, dvg, rts, xtb, vProb, bound, gOnset, dx, dt):
    ncond, ntrials, ntime = rProb.shape
    tb = ntime * dt
    for i in range(ncond):
        tr = gOnset[i]
        for j in range(ntrials):
            ix = sim_dpm_trace_upper(rProb[i,j,tr:], dvg[i,j,tr:], xtb[i], vProb[i], bound[i], dx[i])
            if ix<0:
                rt_ix = ntime + 1
                rts[i,j] = 1000.
            else:
                rt_ix = tr + ix
                rts[i,j] = rt_ix * dt
    #
    # gacc = np.mean(ufunc_where(rts < tb, 1, 0), axis=1)
    # rts[rts>=tb] = 1000.
    # cq = self.RTQ(zip(rts, [self.tb] * nl))
    # return hs([hs([i[ii] for i in [gacc, cq]]) for ii in range(nl)])



# @jit((float64[:,:,:], float64[:,:,:], float64[:,:], float64[:,:], float64[:], float64[:], int64[:], float64[:], float64), nopython=True)
# def sim_many_single(rProb, dvg, rts, xtb, vProb, bound, gOnset, dx, dt):
    # ix = sim_dpm_trace_upper(rProb[i,j,tr:], dvg[i,j,tr:], xtb[i], vProb[i], bound[i], dx[i])
    # if ix<0:
    #     rt_ix = ntime + 1
    #     rts[i,j] = 1000.
    # else:
    #     rt_ix = tr + ix
    #     rts[i,j] = rt_ix * dt

# @jit(nopython=True)
# def sim_multi_radd(results, xtb, idxArray, vProb, bound, onset, B, dx, dt, tb, maxTrials):
#     TargetRT = .52
#     onsetIX = np.int(onset)
#     rProb = randsample((maxTrials, nAlt, 680))
#     for idx in idxArray:
#         idxResults = results[results[:, 0]==idx]
#         ntrials = idxResults.shape[0]
#         vTrial = vProb
#         for t in range(ntrials):
#             vPrevious = vTrial
#             response, rt = sim_multi_go(rProb[t], xtb, vTrial, aTrial, onsetIX, dx, dt)
#             if response:
#                 correct = 1.
#                 score = np.exp(-abs(TargetRT-rt)*9)
#             else:
#                 correct = 0.
#                 score = -1.
#             vTrial = vTrial + (B * (vTrial - vTrial * np.exp((TargetRT-rt))))
#         if vTrial > 1.0:
#             vTrial = .95
#         elif vTrial < .05:
#             vTrial = .1
#         idxResults[t] = np.array([idx, response, correct, rt, score, vTrial])
#     results[results[:, 0]==idx] = idxResults



# if TargetRT <= .52:
#     TargetRT = .52
# elif TargetRT >= tb:
#     TargetRT = tb


# X = xBase * t**-.1
# idxSSD = np.zeros(idxResults[idxResults[:, 2]<1.].shape[0])
# idxSSD = results[results[:, 2]<1.]
# decay = np.exp(np.arange(ntrials)*-xBase)
# ssExpected = .5
# xBase = .0005
# ssdMS = ssdIX*.001
# delta = ssdMS - ssExpected
# ssExpected = ssExpected + C * delta
# lrDecay = (t/10)**-xBase #np.exp(t*-xBase)
# C = lrDecay * C

# TargetRT = TargetRT + ((C * t**-.1) * TargetRT)
# TargetRT = TargetRT + ((C * t**-.1) * TargetRT)
# TargetRT = TargetRT + ((R * t**-.1) * TargetRT)
# TargetRT = TargetRT + .001
# TargetRT = TargetRT + .001 * decay[t]
# TargetRT = .52 + xBase * np.exp(-errT) #(lrDecay * C) * TargetRT
# TargetRT = .52 + C * np.exp(-errT)

# TargetRT = TargetRT - ((R * t**-.1) * TargetRT)
# TargetRT = TargetRT - .001 #* sensitivity#(sensitivity * .01 * TargetRT)
# TargetRT = TargetRT - .001 * decay[t]#(lrDecay * C) * TargetRT
# TargetRT = TargetRT - ((C * t**-.1) * TargetRT)

# TargetRT = .52 + xBase * np.exp(-errT)


# TargetRT = .52 + C * np.exp(-errT)
# * np.exp((stopTargetRT-rt)))
#
# if nthStop<15:
#     lrBound = .1 * C
# else:
#     # lrBound = lrBound + C * (np.std(idxSSD[nthStop-15 : nthStop]) - lrBound)
#     lrBound = C * np.std(idxSSD[nthStop-15 : nthStop])
# aTrial = bound + lrBound * np.exp(-errT)
# aTrial = bound * np.exp(C * delta)
# ssbound = C * np.exp(-errT)
# nthStop += 1

# TargetRT = TargetRT - .001 * sensitivity
# TargetRT = TargetRT - ((R * t**-.1) * TargetRT)
# TargetRT = TargetRT - ((R * t**-.1) * TargetRT)
# if TargetRT <= .52:
#     TargetRT = .52
# elif TargetRT >= tb:
#     TargetRT = tb








# stopTargetRT = stopTargetRT + .002 * stopTargetRT
# if stopTargetRT > tb:
#     stopTargetRT = tb
# TargetRT = TargetRT + C * TargetRT


# stopTargetRT = stopTargetRT - C * stopTargetRT
# if stopTargetRT < TargetRT:
#     stopTargetRT = TargetRT



# ssdVector = idxResults[:,2].std()
# ssdTrue = ssdVector[ssdVector>0].mean()
# lnlike=1.
# lnlike = get_norm(ssExpected, .05, ssOnset*.001)
# ssExpected = ssExpected + C * (ssOnset*.001 - ssExpected)

# vTrial = vTrial * np.exp(-(TargetRT-rt)* B * Learn)
# vTrial = vTrial * np.exp((tb-rt) * B * Learn *
# vTrial = vTrial + (B * Learn * (vTrial - vTrial * np.exp(TargetRT-rt)))
# vTrial = vTrial + (B * Learn * (vTrial - vTrial * np.exp((TargetRT-rt))))
# vTrial = vTrial + (B * Learn * (vTrial - vTrial * np.exp((ssExpected-ssExPrevious))))
# vTrial  np.exp(-(ssOnset*.001 - ssExpected))
# vTrial = vTrial + (C * Learn * (vTrial - vTrial * np.exp((tb-rt))))
# vTrial = vTrial + (B * Learn * (vTrial - vTrial * np.exp((stopTargetRT-rt))))
# vTrial = vTrial + (B * Learn * (vTrial - vTrial * np.exp((stopTargetRT-rt))))

# vsTrial = vsTrial + (C * Learn * (vsTrial - vsTrial * np.exp((stopTargetRT-rt))))

# aTrial = aTrial + C * startLearn * 1/np.exp((ssExpected-ssExPrevious))
# aTrial = aTrial - (C * Learn * (aTrial - aTrial * np.exp((stopTargetRT-rt))))
# aTrial = aTrial + C * Learn * bound
# aTrial = aTrial - (C * Learn * (aTrial - aTrial * np.exp((stopTargetRT-rt))))
# aTrial = aTrial - (C * Learn * (aTrial - aTrial * np.exp((stopTargetRT-rt))))
# aTrial = aTrial + Learn * C * (ssExpected - ssExPrevious)
# aTrial = aTrial + Learn * C * aTrial #np.exp(ssExPrevious - ssExpected)
# aTrial = bound * (t/10)**-C
# aTrial = aTrial * Learn * np.exp(-(stopTargetRT-rt))
# aTrial = aTrial + C * (score - np.mean(idxResults[:t, 6]))
# aTrial = aTrial + C *
# aTrial = aTrial + C * (aTrial - (aTrial * startLearn * np.exp(-(stopTargetRT-rt))))
# aTrial = aTrial + aTrial * Learn * C * np.exp(stopTargetRT-rt) * np.exp(ssOnset*.001 - ssExpected)

# goStartTrial = goStartTrial - (B * startLearn * np.exp((stopTargetRT-rt)))
# goStartTrial = goStartTrial + C * Learn * np.exp((rt-stopTargetRT))
# goStartTrial = goStartTrial - C * startLearn * 1/np.exp((ssExpected-ssExPrevious))

# if goStartTrial > 0:
#     goStartTrial = 0
# if aTrial < .05:
#     aTrial = .06
# elif aTrial > 1.:
#     aTrial = 1.



#!/usr/local/bin/env python
from __future__ import division
import sys
from future.utils import listvalues
from copy import deepcopy
import pandas as pd
import numpy as np
import matplotlib.pyplot as plt
import seaborn as sns
from radd.tools import colors, messages, analyze
from radd.tools.analyze import bootstrap_data
from radd import theta
from IPython.display import display, Latex
from ipywidgets import IntProgress, HTML, Box
from itertools import product


def pandaify_results(gort, ssrt, tb=.68, clmap=None, bootstrap=False, bootinfo={'nsubjects':25, 'ntrials':1000, 'groups':['ssd']}, ssd=np.array([[.2, .25, .3, .35, .4]])):


    nlevels = gort.shape[0]
    if nlevels==1:
        clmap = {'flat': ['flat']}
    conds = list(clmap)
    cond_levels = list(product(*[list(levels) for levels in listvalues(clmap)]))
    dfColumns=['ttype', 'ssd', 'response', 'acc', 'rt', 'ssrt', 'trial']
    dfColumns = conds + dfColumns

    # bootinfo['groups'] = bootinfo['groups'] + conds
    dfList = []
    for i in range(nlevels):
        df = create_sim_dataframe(gort[i], ssrt[i], tb=tb, ssd=ssd)
        for ii, c in enumerate(conds):
            df[c] = cond_levels[i][ii]
        dfList.append(df[dfColumns])

    resultsdf = pd.concat(dfList)
    resultsdf.reset_index(drop=True, inplace=True)

    if bootstrap:
        resultsdf = bootstrap_data(resultsdf, nsubjects=bootinfo['nsubjects'], n=bootinfo['ntrials'], groups=bootinfo['groups'])

    return resultsdf


def create_sim_dataframe(rt, ssrt, tb=.68, ssd=np.array([.2, .25, .3, .35, .4])):

    ntrials = rt.shape[0]
    nssd, nssper = ssrt.shape
    nss = ssrt.size
    ssd = ssd.squeeze()
    ttype = ['go']*ntrials
    ttype[:nss] = ['stop']*nss
    trials = np.arange(ntrials)

    rts = rt.ravel()
    ssrts = ssrt.reshape(nss)
    ssrts = np.concatenate((ssrts, np.array([np.nan]*nss)))
    ssds = np.hstack([np.hstack(np.sort(np.tile(ssd, nssper))), np.ones(ntrials-nss)])
    ssds = (np.round(ssds, 2) * 1000).astype(int)
    df = pd.DataFrame({'ttype':ttype, 'rt': rts, 'ssrt':ssrts, 'trial':trials, 'ssd':ssds})

    df['response'] = np.nan
    df['acc'] = np.nan
    stopdf, godf = [df[df.ttype==ttype] for ttype in ['stop', 'go']]
    godf.loc[godf.rt < tb, ['response', 'acc']] = 1.
    godf.loc[godf.rt >= tb, ['response', 'acc']] = 0.
    stopdf.loc[stopdf.rt < stopdf.ssrt, 'response'] = 1
    stopdf.loc[stopdf.rt >= stopdf.ssrt, 'response'] = 0
    stopdf.loc[stopdf.response==1, 'acc'] = 0
    stopdf.loc[stopdf.response==0, 'acc'] = 1
    return pd.concat([stopdf, godf])



class PBinJ(object):
    """ initialize multiple progress bars for tracking nested stages of fitting routine
    """
    def __init__(self, n=1, value=0, status='{}', color='b', width='60%', height='22px'):
        self.displayed = False
        self.style_bar(n=n, value=value, status=status, color=color, width=width, height=height)

    def style_bar(self, n=1, value=0, status='{}', color='b', width='60%', height='22px'):
        colordict = {'g': 'success', 'b': '', 'r': 'danger', 'y': 'warning', 'c': 'info'}
        bar_style = colordict[color]
        self.bar = IntProgress(min=0, max=n, value=value, bar_style=bar_style)
        self.status = status
        self.bar.bar_style = bar_style
        self.bar.width = width
        self.bar.height = height

    def reset_bar(self, color=False):
        self.update(value=0)

    def update(self, value=None, status=None):
        if not self.displayed:
            display(self.bar)
            self.displayed=True
        if status is not None:
            if hasattr(status, '__iter__') and not isinstance(status, str):
                status = self.status.format(*status)
            else:
                status = self.status.format(status)
            self.bar.description = status
        if value is not None:
            self.bar.value = value+1

    def clear(self):
        self.bar.close()
        self.displayed = False


class GlobalCallback(object):
    """ A callback function for reporting basinhopping status
    Arguments:
        x (array):
            parameter values
        fmin (float):
            function value of the trial minimum, and
        accept (bool):
            whether or not that minimum was accepted
    """
    def __init__(self,  n=1, value=0, status='{:.5fz} / {:.5fz}', color='r', fmin=1000., method='basin'):

        if method=='basin':
            self.callback = self.basinhopping_callback
        else:
            self.callback = self.evolution_callback
            fmin = 1000.
        self.pbar = PBinJ(n=n, value=value, status=status, color=color)
        self.reset(history=True, gbasin=True, fmin=fmin)
        self.xhistory = []
        # status=(MyFloat(x) for x in [fmin, fmin])
        # self.pbar.update(value=0., status=status)

    def reset(self, history=True, bar=False, gbasin=False, get_call=False, fmin=1000.):
        fmin = MyFloat(fmin)
        if history:
            self.history = [fmin]
        if gbasin:
            self.gbasin = fmin
        if bar:
            self.pbar.reset_bar()
        if get_call:
            return self.callback

    def basinhopping_callback(self, x, fmin, accept):
        if fmin <= np.min(self.history) and fmin<=self.gbasin:
            self.gbasin = fmin
            self.xhistory.append((fmin, x))
            self.reset(history=True, bar=True)
        if accept:
            self.history.append(fmin)
            status=(MyFloat(x) for x in [self.gbasin, fmin])
            self.pbar.update(value=len(self.history), status=status)

    def evolution_callback(self, xk, convergence):
        #  convergence >= np.max(self.history) and
        if (1-convergence)<self.gbasin:
            self.gbasin = 1-convergence
            self.xhistory.append((1-convergence, xk))
            # print(self.xhistory[-1])
            # self.reset(history=True, bar=True, fmin=0)

        self.history.append(1-convergence)
        status=(MyFloat(x) for x in [self.gbasin, 1-convergence])
        self.pbar.update(value=len(self.history), status=status)

    def clear(self):
        self.pbar.clear()


class GradientCallback(object):
    """ A callback function for reporting basinhopping status
    Arguments:
        x (array):
            parameter values
        fmin (float):
            function value of the trial minimum, and
        accept (bool):
            whether or not that minimum was accepted
    """
    def __init__(self,  n=1, value=0, status='{:.5fz} / {:.5fz}', color='g', fmin=1000.):
        self.pbar = PBinJ(n=n, value=value, status=status, color=color)
        self.reset(history=True, lbasin=True, fmin=fmin)
        self.xhistory = []
        self.nsame = 0
        # status=(MyFloat(x) for x in [self.lbasin, fmin])
        # self.pbar.update(value=0., status=status)

    def reset(self, history=True, bar=False, lbasin=False, get_call=False, fmin=1000.):
        fmin = MyFloat(fmin)
        if history:
            self.history = [fmin]
            self.nsame = 0
        if lbasin:
            self.lbasin = fmin
        if bar:
            self.pbar.reset_bar()
        if get_call:
            return self.callback

    def callback(self, params, iter, resid):
        fmin = np.sum(resid**2)
        if fmin < np.min(self.history) and fmin<self.lbasin:
            self.history.append(fmin)
            self.lbasin = fmin
            self.xhistory.append((fmin, params.valuesdict()))
            self.nsame = 0
        else:
            self.nsame += 1
            #self.reset(history=True, bar=True)
        status=(MyFloat(x) for x in [self.lbasin, fmin])
        self.pbar.update(value=iter, status=status)

    def clear(self):
        self.pbar.clear()



class MyFloat(float):
    """ remove leading zeros from string formatted floats
    """
    def remove_leading_zero(self, value, string):
        if 1 > value > -1:
            string = string.replace('0', '', 1)
        return string

    def __format__(self, format_string):
        if format_string.endswith('z'):
            format_string = format_string[:-1]
            removezero = True
        else:
            removezero = False
        string = super(MyFloat, self).__format__(format_string)
        return self.remove_leading_zero(self, string) if removezero else string


def rwr(X, get_index=False, n=None):
    """
    Modified from http://nbviewer.ipython.org/gist/aflaxman/6871948
    """
    if isinstance(X, pd.Series):
        X = X.copy()
        X.index = range(len(X.index))
    if n == None:
        n = len(X)
    resample_i = np.floor(np.random.rand(n) * len(X)).astype(int)
    X_resample = (X[resample_i])
    if get_index:
        return resample_i
    else:
        return X_resample


def resample_data(data, n=120, groups=['ssd']):
    """ generates n resampled datasets using rwr()
    for bootstrapping model fits
    """
    df = data.copy()
    bootlist = list()
    if n == None:
        n = len(df)
    for level, level_df in df.groupby(groups):
        boots = level_df.reset_index(drop=True)
        orig_ix = np.asarray(boots.index[:])
        resampled_ix = rwr(orig_ix, get_index=True, n=n)
        bootdf = level_df.irow(resampled_ix)
        bootlist.append(bootdf)
    # concatenate and return all resampled conditions
    return self.model.rangl_data(pd.concat(bootlist))


def extract_popt_fitinfo(finfo=None, plist=None, pcmap=None):
    """ takes optimized dict or DF of vectorized parameters and
    returns dict with only depends_on.keys() containing vectorized vals.
    Is accessed by fit.Optimizer objects after optimization routine.
    ::Arguments::
    finfo (dict/DF):
        finfo is dict if self.fit_on is 'average'
        and DF if self.fit_on is 'subjects' or 'bootstrap'
        contains optimized parameters
    ::Returns::
    popt (dict):
        dict with only depends_on.keys() containing
        vectorized vals
    """
    finfo = dict(deepcopy(finfo))
    plist = list(inits)
    popt = {pkey: finfo[pkey] for pkey in plist}
    for pkey in list(pcmap):
        popt[pkey] = np.array([finfo[pc] for pc in pcmap[pkey]])
    return popt


def params_io(p={}, io='w', iostr='popt'):
    """ read // write parameters dictionaries
    """
    if io == 'w':
        pd.Series(p).to_csv(''.join([iostr, '.csv']))
    elif io == 'r':
        ps = pd.read_csv(''.join([iostr, '.csv']), header=None)
        p = dict(zip(ps[0], ps[1]))
        return p


def fits_io(fitparams, fits=[], io='w', iostr='fits'):
    """ read // write y, wts, yhat arrays
    """
    y = fitparams['y'].flatten()
    wts = fitparams['wts'].flatten()
    fits = fits.flatten()
    if io == 'w':
        index = np.arange(len(fits))
        df = pd.DataFrame({'y': y, 'wts': wts, 'yhat': fits}, index=index)
        df.to_csv(''.join([iostr, '.csv']))
    elif io == 'r':
        df = pd.read_csv(''.join([iostr, '.csv']), index_col=0)
        return df



from __future__ import division
from future.utils import listvalues
from copy import deepcopy
import pandas as pd
import numpy as np
from numpy.random import random_sample as randsample
from scipy.stats.mstats import mquantiles
from itertools import product
from radd import build, theta
from radd.tools.analyze import blockify_trials
from radd.compiled import jitfx
import itertools

class Simulator(object):
    def __init__(self, inits, data=None, fitparams=None, constants=[], ssdMethod='all', nblocks=25, nruns=10):
        self.analyzeProbes = False
        self.ssdMethod = ssdMethod
        self.kind = fitparams.kind
        self.quantiles = fitparams.quantiles
        self.fitparams = fitparams
        self.nruns = nruns
        self.allparams = ['B', 'C', 'R', 'a', 'ssv', 'tr', 'v', 'xb']
        self.update(fitparams=self.fitparams, inits=inits, data=data, constants=constants, nblocks=nblocks)


    def update(self, **kwargs):
        kw_keys = list(kwargs)
        self.tb = self.fitparams['tb']
        self.dt = self.fitparams['dt']
        self.si = self.fitparams['si']
        if 'fitparams' in kw_keys:
            self.fitparams = kwargs['fitparams']
        if 'inits' in kw_keys:
            self.inits = deepcopy(kwargs['inits'])
            if 'B' not in list(self.inits):
                self.inits['B'] = .1
                self.inits['C'] = .002
                self.inits['R'] = .1
            self.theta = pd.Series(self.inits)
        if 'nruns' in kw_keys:
            self.nruns = kwargs['nruns']
        if 'constants' in kw_keys:
            const = kwargs['constants']
            self.constants = [p for p in self.allparams if p in const]
        if 'nblocks' in kw_keys:
            self.nblocks = kwargs['nblocks']
            self.blocksCol = 'block{}'.format(self.nblocks)
            if hasattr(self, 'data'):
                self.data = blockify_trials(self.data, nblocks=self.nblocks)
                self.rtBlocks, rtErr, self.saccBlocks, saccErr = self.blockify_data(self.data, measures=['rt', 'acc'], get_var=True)
                self.rt_weights = rtErr.mean() / rtErr
                self.sacc_weights = saccErr.mean() / saccErr
        if 'data' in kw_keys:
            self.get_trials_data(kwargs['data'])
        self.nlevels = self.fitparams['nlevels']
        self.y = self.fitparams.y.flatten()
        self.wts = self.fitparams.wts.flatten()
        self.format_params()
        self.make_results_matrix()
        self.make_io_vectors()


    def cost_fx_rl(self, p):
        resultsDF = self.simulate_model(p, analyze=False)
        return self.analyze_trials(resultsDF)


    def cost_fx(self, p, analyze=True):
        resultsDF = self.simulate_model(p, analyze=False)
        return self.analyze_trials(resultsDF)


    def cost_fx_lmfit(self, lmParams, sse=False):
        # resultsDF = self.simulate_model_nruns(lmParams.valuesdict(), analyze=False)
        resultsDF = self.simulate_model(lmParams.valuesdict(), analyze=False)
        saccBlocks = resultsDF[resultsDF.ttype==0].groupby(self.blocksCol).mean().acc.values
        rtBlocks = resultsDF[resultsDF.response==1].groupby(self.blocksCol).mean().rt.values
        rtErr = self.rt_weights * (rtBlocks*15 - self.rtBlocks*15)
        saccErr = self.sacc_weights * (saccBlocks - self.saccBlocks)
        return np.hstack([rtErr, saccErr])


    # def simulate_model_nruns(self, p, analyze=True):
        #     self.preproc_params(p)
        #     res = jitfx.sim_dpm_learning_nruns(self.nresults, self.rProb, self.rProbSS, self.xtb, self.idxArray, self.vProb, self.vsProb, self.bound, self.gOnset, self.Beta, self.Alpha, self.Rho, self.dx, self.dt, self.tb, self.ntrials, self.nruns)
        #     if analyze:
        #         return self.analyze(res)
        #     self.nresultsDF.loc[:, self.resultsHeader] = res
        #     return self.nresultsDF

    # def simulate_model(self, p, analyze=True):
    #     self.preproc_params(p)
    #     results = np.copy(self.results)
    #     jitfx.sim_dpm_learning(results, self.rProb, self.rProbSS, self.xtb, self.idxArray, self.vProb, self.vsProb, self.bound, self.gOnset, self.Beta, self.Alpha, self.Rho, self.dx, self.dt, self.tb, self.ntrials)
    #     if analyze:
    #         return self.analyze(results)
    #     self.resultsDF.loc[:, self.resultsHeader] = results
    #     return self.resultsDF
        # resultsDF = pd.DataFrame(self.results, columns=self.resultsHeader)
        #pd.concat([resultsDF, self.data[['cond', 'sstrial', 'probe', 'trial', self.blocksCol]]], axis=1)
        # return self.resultsDF

    def simulate_model(self, p, analyze=True):
        self.preproc_params(p)
        results = np.copy(self.results)
        jitfx.sim_dpm_learning(results, self.rProb, self.rProbSS, self.xtb, self.idxArray, self.drift, self.ssdrift, self.bound, self.gOnset, self.Beta, self.Alpha, self.Rho, self.dx, self.dt, self.tb, self.ntrials)
        if analyze:
            return self.analyze(results)
        self.resultsDF.loc[:, self.resultsHeader] = results
        return self.resultsDF


    def analyze(self, res):
        idxIX, ttypeIX, ssdIX, respIX, accIX, rtIX, scoreIX = self.dataColsIX
        # calc go and stop accuracy
        goTrials = res[res[:, ttypeIX]==1.]
        ssTrials = res[res[:, ttypeIX]==0.]
        gAcc = goTrials[:, accIX].mean()
        if self.analyzeProbes:
            if self.ssdMethod=='all':
                ssdAccuracy = [res[res[:, ssdIX]==ssd][:, accIX] for ssd in self.probes]
                sAcc = np.array([np.nanmean(ssdAcc) for ssdAcc in ssdAccuracy])
            else:
                sAcc = np.nanmean(np.hstack(ssdAccuracy))
        else:
            sAcc = ssTrials[:, accIX].mean()
        # calc correct and error rt quantiles
        respTrials = res[res[:, respIX]==1.]
        corRT = respTrials[respTrials[:, accIX]==1.][:, rtIX]
        errRT = respTrials[respTrials[:, accIX]==0.][:, rtIX]
        cq = mquantiles(corRT[corRT < self.tb], self.quantiles)
        eq = mquantiles(errRT[errRT < self.tb], self.quantiles)
        # concatenate into single cost vector
        return np.hstack([gAcc, sAcc, cq, eq])


    def preproc_params(self, p, asarray=False):

        params = deepcopy(self.fixedParams)
        if isinstance(p, np.ndarray):
            p = dict(zip(self.pvary, p))
        params.update(p)

        try:
            self.Beta = params['B']
            self.Alpha = params['C']
            self.Rho = params['R']
        except Exception:
            pass

        self.bound = params['a']
        if 'xb' in self.modelparams:
            self.xtb = np.cosh(params['xb'] * self.xtime).squeeze()
        else:
            self.xtb = np.ones(self.xtime.size)
        self.vProb = 0.5 * (1 + (params['v'] * np.sqrt(self.dt))/self.si)
        self.vsProb = 0.5 * (1 + (params['ssv'] * np.sqrt(self.dt))/self.si)
        self.drift = params['v']
        self.ssdrift = params['ssv']
        self.gOnset = params['tr'] / self.dt
        self.gonsetTime = params['tr']

        if asarray:
            pSeries = pd.Series(params)[self.pvary]
            return pSeries.values


    def make_io_vectors(self):
        self.ntrials = self.data.groupby('idx').count()['trial'].max()
        # generate vectors of random floats [0-1)
        self.rProb = randsample((self.ntrials, self.ntime))
        self.rProbSS = randsample((self.ntrials, self.ntime))


    def make_results_matrix(self):

        dataCols = ['idx', 'ttype', 'ssd', 'response', 'acc', 'rt']
        index = self.data.index.values
        resCols = np.array(dataCols + ['score', 'drift', 'bound'])
        resData = np.zeros((index.size, resCols.size))
        resultsDF = pd.DataFrame(resData, columns=resCols, index=index)
        nidx = self.data.idx.unique().size

        resultsDF.loc[:, ['idx', 'ttype', 'ssd']] = self.data.loc[:, ['idx', 'ttype', 'ssd']]

        self.resultsHeader = resultsDF.columns.tolist()
        self.results = resultsDF.values
        self.resultsDF = pd.concat([resultsDF, self.data[['cond', 'sstrial', 'probe', 'trial', self.blocksCol]]], axis=1)

        self.dataColsIX = [self.resultsHeader.index(col) for col in dataCols + ['score']]
        self.rtMatrix = np.zeros((nidx, self.nblocks))
        self.saccMatrix = np.zeros((nidx, self.nblocks))
        self.nresultsDF = pd.concat([self.resultsDF]*self.nruns)
        self.nresultsDF.reset_index(inplace=True, drop=True)
        self.nresults = np.vstack([self.results]*self.nruns)
        #
        # dataCols = ['idx', 'ttype', 'ssd', 'response', 'acc', 'rt']
        # index = self.data.index.values
        # resCols = np.array(dataCols + ['alpha', 'drift', 'bound', 'idxResp', 'idxRT'])
        # resData = np.zeros((index.size, resCols.size))
        # resultsDF = pd.DataFrame(resData, columns=resCols, index=index)
        # nidx = self.data.idx.unique().size
        #
        # resultsDF.loc[:, ['idx', 'ttype', 'ssd', 'idxResp', 'idxRT']] = self.data.loc[:, ['idx', 'ttype', 'ssd', 'response', 'rt']].values
        #
        # self.resultsHeader = resultsDF.columns.tolist()
        # self.results = resultsDF.values
        # self.resultsDF = pd.concat([resultsDF, self.data[['cond', 'sstrial', 'probe', 'trial', self.blocksCol]]], axis=1)
        #
        # self.dataColsIX = [self.resultsHeader.index(col) for col in dataCols + ['alpha']]
        # self.rtMatrix = np.zeros((nidx, self.nblocks))
        # self.saccMatrix = np.zeros((nidx, self.nblocks))
        # self.nresultsDF = pd.concat([self.resultsDF]*self.nruns)
        # self.nresultsDF.reset_index(inplace=True, drop=True)
        # self.nresults = np.vstack([self.results]*self.nruns)


    def get_trials_data(self, data):
        if 'probe' in data.columns:
            probeDF = data[data.probe==1]
            self.probes = np.sort(.001*probeDF.ssd.unique() / self.dt)
            self.analyzeProbes = True
        data = data.copy()
        data.loc[:, 'ssd'] = .001*data.ssd.values / self.dt
        data.replace({'ttype': {'go':1., 'stop':0.}}, inplace=True)
        data = data.reset_index(drop=True)
        self.data = blockify_trials(data, nblocks=self.nblocks)
        self.rtBlocks, rtErr, self.saccBlocks, saccErr = self.blockify_data(self.data, measures=['rt', 'acc'], get_var=True)

        self.rt_weights = rtErr.mean() / rtErr
        self.sacc_weights = saccErr.mean() / saccErr
        self.idxArray = self.data.idx.unique()


    def format_params(self):
        self.dx = self.si * np.sqrt(self.dt)
        # self.dx = np.sqrt(self.si * self.dt)
        self.ntime = int(self.tb / self.dt)
        self.xtime = np.cumsum([self.dt] * self.ntime)
        self.modelparams = [p for p in self.allparams if p in list(self.inits)]
        self.pflat = [p for p in self.modelparams if p in self.constants]
        self.pvary  = [p for p in self.modelparams if p not in self.pflat]
        self.fixedParams = self.theta[self.pflat].to_dict()
        if 'xb' not in self.modelparams:
            self.fixedParams['xb'] = 0.
        # number of cells in condition matrix (df index)
        self.nvary = np.ones(len(self.pvary)).astype(np.int64)
        self.preproc_params(self.inits)


    def make_empircal_data_matrix(self):
        ttypeIX, ssdIX, respIX, accIX, rtIX = self.dataColsIX
        res = self.data
        goTrials = res[res[:, ttypeIX]==1.]
        ssTrials = res[res[:, ttypeIX]==0.]
        self.GoAccEmp = goTrials[:, accIX]
        self.ssAccEmp = ssTrials[:, accIX]
        self.CorRTEmp = goTrials[goTrials[:, accIX]==1.][:, rtIX]
        self.ErrRTEmp = goTrials[goTrials[:, accIX]==0.][:, rtIX]
        respTrials = res[res[:, respIX]==1.]
        self.CorRTEmp = respTrials[respTrials[:, accIX]==1.][:, rtIX]
        self.ErrRTEmp = respTrials[respTrials[:, accIX]==0.][:, rtIX]


    def blockify_data(self, data, get_var=False, measures=['rt', 'acc']):
        data = data.copy()
        if self.blocksCol not in data.columns:
            data = blockify_trials(data, nblocks=self.nblocks)
        goDF = data[data.response==1.]
        ssDF = data[data.ttype==0.]
        tableList = []
        if 'rt' in measures:
            rtTable = pd.pivot_table(goDF, values='rt', columns=self.blocksCol, index='idx')
            tableList.append(rtTable)
        if 'acc' in measures:
            ssTable = pd.pivot_table(ssDF, values='acc', columns=self.blocksCol, index='idx')
            tableList.append(ssTable)
        if 'bound' in measures:
            scoreTable = pd.pivot_table(data, values='bound', columns=self.blocksCol, index='idx')
            tableList.append(scoreTable)
        if 'score' in measures:
            scoreTable = pd.pivot_table(data, values='score', columns=self.blocksCol, index='idx')
            tableList.append(scoreTable)
        if 'vTrial' in measures:
            vTable = pd.pivot_table(data, values='drift', columns=self.blocksCol, index='idx')
            tableList.append(vTable)
        if get_var:
            blockedMeasures = [[table.mean().values, table.sem().values*1.96] for table in tableList]
            blockedMeasures = list(itertools.chain.from_iterable(blockedMeasures))
        else:
            blockedMeasures = [table.mean().values for table in tableList]
        return blockedMeasures


    def analyze_trials(self, resultsDF):
        saccBlocks = resultsDF[resultsDF.ttype==0].groupby(self.blocksCol).mean().acc.values
        rtBlocks = resultsDF[resultsDF.response==1].groupby(self.blocksCol).mean().rt.values
        rtErr = np.sum((self.rt_weights * (rtBlocks*10 - self.rtBlocks*10))**2)
        saccErr = np.sum((self.sacc_weights * (saccBlocks - self.saccBlocks))**2)
        return rtErr + saccErr



#!usr/bin/env python
from __future__ import division
import os
from future.utils import listvalues
from copy import deepcopy
import pandas as pd
import numpy as np
from numpy import array
from radd.tools import analyze
from radd import theta
from itertools import product


class DataHandler(object):

    def __init__(self, model, max_wt=3., verbose=False):
        self.model = model
        self.data = model.data
        self.inits = model.inits
        self.model_id = model.model_id
        self.idx = model.idx
        self.nidx = model.nidx
        self.weighted = model.weighted
        self.max_wt = max_wt
        self.ssd_method = model.ssd_method
        self.kind = model.kind
        self.fit_on = model.fit_on
        self.quantiles = model.quantiles
        self.groups = model.groups
        self.depends_on = model.depends_on
        self.pcmap = model.pcmap
        self.clmap = model.clmap
        self.conds = model.conds
        self.nconds = model.nconds
        self.nlevels = model.nlevels
        self.cond_matrix = model.cond_matrix
        self.bwfactors = model.bwfactors
        self.nrows = self.nidx * model.nlevels
        self.grpData = self.data.groupby(self.groups)
        self.verbose = verbose


    def make_dataframes(self):
        """ Generates the following dataframes and arrays:
        observed (list of ndarrays):
            list containing ndarrays (ncond x n_observed_data_points) entered into thec
            costfx during fitting. If self.fit_on=='average', self.observed will
            contain a single ndarray (ncond x average_data). If self.fit_on=='subjects',
            self.observed will contain 1 ndarray for each subject and fits will be performed
            iteratively
        observedDF (DF):
              Contains Prob and RT quant. for each subject
              used to calc. cost fx weights
        fits (DF):
              empty DF shaped like observed DF, used to store simulated
              predictions of the optimized model
        fitinfo (DF):
              stores all opt. parameter values and model fit statistics
        """
        # make observed y and wts dataframes (all subjects)
        self.make_observed_groupDFs()
        # make yhatdf to fill w/ model-predicted data arrays
        self.yhatdf = self.make_yhat_df()
        # make fitdf for storing w/ goodness-of-fit stats and popt
        self.fitdf = self.make_fit_df()
        # make poptdf for storing popt of conditional models as matrix
        self.poptdf = self.make_popt_df()
        odf = self.observedDF.copy()
        wdf = self.wtsDF.copy()
        condvalues = lambda df: df.loc[:, 'acc':].dropna(axis=1).values.squeeze()
        flatvalues = lambda df: df.loc['acc':].values.squeeze()
        if self.fit_on=='subjects':
            self.observed = [condvalues(odf[odf['idx']==idx]) for idx in self.idx]
            self.cond_wts = [condvalues(wdf[wdf['idx']==idx]) for idx in self.idx]
            self.observed_flat = [flatvalues(odf[odf['idx']==idx].mean()) for idx in self.idx]
            self.flat_wts = [flatvalues(wdf[wdf['idx']==idx].mean()) for idx in self.idx]
        elif self.fit_on=='average':
            self.observed = [condvalues(odf.groupby(self.conds).mean())]
            self.cond_wts = [condvalues(wdf.groupby(self.conds).mean())]
            self.observed_flat = [flatvalues(odf.mean())]
            self.flat_wts = [flatvalues(wdf.mean())]


    def make_observed_groupDFs(self):
        """ concatenate all idx data vectors into a dataframe
        """
        odf_header = self.make_headers()
        data = self.data.copy()
        ssdmethod = self.ssd_method
        self.grpData = data.groupby(np.hstack(['idx', self.conds]).tolist())
        datdf = self.grpData.apply(analyze.rangl_data, ssdmethod, self.quantiles).sortlevel(0)
        # self.datdf = datdf
        groupvalues = datdf.reset_index()[self.groups].values
        nan_data = np.zeros((groupvalues.shape[0], len(odf_header)), dtype=np.int64)
        self.observedDF = pd.DataFrame(nan_data, columns=odf_header, index=range(nan_data.shape[0]))
        self.observedDF.loc[:, self.groups] = groupvalues
        self.wtsDF = self.make_wts_df()
        for rowi in self.observedDF.index.values:
            # fill observedDF one row at a time, using idx_rows
            self.observedDF.loc[rowi, self.idx_cols[rowi]] = datdf.values[rowi]

        if self.bwfactors is not None and self.model.fit_on=='subjects':
            bwix = self.observedDF[self.groups].columns.size
            bwunique = [self.data[self.data.idx==idx][self.bwfactors].unique() for idx in self.idx]
            #wfactors = list(self.clmap)
            #wfactors.remove(self.bwfactors)
            # nwithin = np.sum([len(self.clmap[wfactor]) for wfactor in wfactors])
            bwcast = np.hstack([np.tile(bw, self.nlevels) for bw in bwunique])

            print(len(bwcast))
            print(self.observedDF.shape[0])
            print(self.bwfactors)
            #print(self.observedDF)
            self.observedDF.insert(bwix, self.bwfactors, bwcast)
            self.wtsDF.insert(bwix, self.bwfactors, bwcast)
            errdf = self.observedDF.groupby(self.conds+[self.bwfactors]).sem()*2.
            self.observedErr = errdf.reset_index()[self.observedDF.columns[1:]]
        else:
            errdf = self.observedDF.groupby(self.conds).sem()*2
            self.observedErr = errdf.reset_index()[self.observedDF.columns[1:]]


    def make_freq_df(self):
        data = self.data.copy()
        self.grpData = data.groupby(self.groups)
        freqdf = self.grpData.apply(analyze.rangl_freq, self.quantiles)
        countdf =  self.grpData.apply(analyze.rangl_counts)

        metadf = freqdf.reset_index()[self.groups]
        bins = np.arange(self.quantiles.size+1)
        freqcols = sum([['{}{}'.format(rtype, i) for i in bins+1] for rtype in ['o', 'e']], [])
        freqvals = pd.DataFrame(np.vstack(freqdf.values), columns=freqcols)
        countvals = pd.DataFrame(np.vstack(countdf.values), columns=['Ncor', 'Nerr'])
        count_freq_vals = pd.concat([countvals, freqvals], axis=1)
        freqDF = pd.concat([metadf, count_freq_vals], axis=1)
        return freqDF


    def make_wts_df(self):
        """ calculate and store cost_function weights
        for all subjects/conditions in data
        """
        wtsDF = self.observedDF.copy()
        if self.weighted:
            try:
                # calc & fill wtsDF with idx quantile and accuracy weights (ratios)
                quant_wts, acc_wts = self.calc_empirical_weights()
                qwts = np.vstack(quant_wts).reshape(wtsDF.shape[0], -1)
                awts = np.vstack(acc_wts).reshape(wtsDF.shape[0], -1)
                wtsDF.loc[:, self.q_cols] = qwts
                wtsDF.loc[:, self.p_cols] = awts
                wts_numeric = wtsDF.loc[:, 'acc':]
                wtsDF.loc[:, 'acc':] = wts_numeric.apply(analyze.fill_nan_vals, axis=1)
            except Exception:
                if self.verbose:
                    print("Unable to calculate cost f(x) weights, setting all w=1.")
                wtsDF.loc[:, self.p_cols+self.q_cols] = 1.
        else:
            wtsDF.loc[:, self.p_cols+self.q_cols] = 1.
        return wtsDF.copy()


    def calc_empirical_weights(self):
        """ calculates weight vectors for observed correct & err RT quantiles and
        go and stop accuracy for each subject (see funcs in radd.tools.analyze)
        """
        data = self.data.copy()
        # quant_wts = [analyze.idx_quant_weights(df, conds=self.conds, max_wt=self.max_wt, quantiles=self.quantiles, bwfactors=self.bwfactors) for i, df in data.groupby('idx')]
        quant_wts = analyze.idx_quant_weights_OLD(data, prob=self.quantiles, groups=self.groups, nsplits=np.cumprod(self.cond_matrix)[-1], max_wt=self.max_wt)
        acc_wts = [analyze.idx_acc_weights(df, conds=self.conds, ssd_method=self.ssd_method) for i, df in data.groupby('idx')]
        return quant_wts, acc_wts


    def get_cond_combos(self):
        clevels = [list(self.clmap[c]) for c in np.sort(list(self.clmap))]
        level_data = list(product(*clevels))
        return pd.DataFrame(level_data, columns=self.groups[1:])


    def make_yhat_df(self):
        """ make empty dataframe for storing model predictions (yhat)
        """
        yhatcols = self.observedDF.columns
        indx = np.arange(self.nlevels)
        yhatdf = pd.DataFrame(np.nan, index=indx, columns=yhatcols)
        minfo = self.get_cond_combos()
        yhatdf.loc[:, minfo.columns] = minfo
        yhatdf = yhatdf.copy()
        yhatdf.insert(len(self.groups), 'pvary', np.nan)
        self.empty_yhatdf = yhatdf.copy()
        return yhatdf


    def make_popt_df(self):
        """ make empty dataframe for storing popt after each fit
        """
        indx = np.arange(self.nlevels)
        poptcols = self.groups + self.poptdf_cols
        poptdf = pd.DataFrame(np.nan, index=indx, columns=poptcols)
        minfo = self.get_cond_combos()
        poptdf[minfo.columns] = minfo
        poptdf.insert(len(self.groups), 'pvary', np.nan)
        self.empty_poptdf = poptdf.copy()
        return poptdf


    def make_fit_df(self):
        """ make empty dataframe for storing fit info
        """
        fitdf = pd.DataFrame(np.nan, index=[0], columns=self.f_cols)
        self.empty_fitdf = fitdf.copy()
        return fitdf


    def fill_poptdf(self, popt, fitparams=None):
        """ fill fitdf with fit statistics
        ::Arguments::
            popt (dict):
                fitinfo Series containing model statistics and
                optimized parameters (see Model.assess_fit() method)
            fitparams (Series):
                model.fitparams dict w/ meta info for last fit
        """

        if fitparams is None:
            fitparams = self.model.fitparams

        poptdf = self.empty_poptdf.copy()
        poptdf['idx'] = str(fitparams.idx)
        poptdf['pvary'] = '_'.join(list(self.model.depends_on))

        p = pd.Series(deepcopy(popt))[self.poptdf_cols].to_dict()
        poptdf.loc[:, self.poptdf_cols] = pd.DataFrame(p, index=poptdf.index)
        # popt = self.model.simulator.vectorize_params(popt)
        # popt_vals = np.array([popt[pkey] for pkey in self.poptdf_cols]).T
        # poptdf.loc[:, self.poptdf_cols] = popt_vals
        if np.any(self.poptdf.isnull()):
            poptdf = poptdf
        else:
            poptdf = pd.concat([self.poptdf, poptdf], axis=0)

        self.poptdf = poptdf.reset_index(drop=True)
        return self.poptdf


    def fill_fitdf(self, finfo, fitparams=None):
        """ fill fitdf with fit statistics
        ::Arguments::
            finfo (Series):
                fitinfo Series containing model statistics and
                optimized parameters (see Model.assess_fit() method)
            fitparams (Series):
                model.fitparams dict w/ meta info for last fit
        """
        if fitparams is None:
            fitparams = self.model.fitparams
        fitdf = self.empty_fitdf.copy()
        pvary = list(self.model.depends_on)

        for fcol in finfo.keys():
            if fcol in pvary:
                pkey = '_'.join([fcol, 'avg'])
                fitdf.loc[0, pkey] = np.mean(finfo[fcol])
                continue
            fitdf.loc[0, fcol] = finfo[fcol]
        if np.any(self.fitdf.isnull()):
            fitdf = fitdf
        else:
            fitdf = pd.concat([self.fitdf, fitdf], axis=0)
        self.fitdf = fitdf.reset_index(drop=True)
        return self.fitdf


    def fill_yhatdf(self, yhat, fitparams=None):
        """ fill yhatdf with model predictions
        ::Arguments::
            yhat (ndarray):
                array containing model predictions (nlevels x ncols)
                where ncols is number of data columns in observedDF
            fitparams (Series):
                model.fitparams dict w/ meta info for last fit
        """
        if fitparams is None:
            fitparams = self.model.fitparams
        nl = fitparams['nlevels']
        yhat = yhat.reshape(nl, -1)
        yhatdf = self.empty_yhatdf.copy()
        yhatdf.loc[:, 'acc':] = yhat
        yhatdf['idx'] = str(fitparams.idx)
        yhatdf['pvary'] = '_'.join(list(self.model.depends_on))
        if np.any(self.yhatdf.isnull()):
            yhatdf = yhatdf
        else:
            yhatdf = pd.concat([self.yhatdf, yhatdf], axis=0)
        self.yhatdf = yhatdf.reset_index(drop=True)
        return self.yhatdf


    def set_model_ssds(self, stopdf, index=['idx']):
        """ set model attr "ssd" as list of np.arrays
        ssds to use when simulating data during optimization
        """
        if self.ssd_method is None:
            self.ssd_method = analyze.determine_ssd_method(stopdf)
            self.model.ssd_method = self.ssd_method

        bwfactors = self.bwfactors
        if bwfactors is not None:
            stop_dfs = stopdf.groupby(bwfactors)
        else:
            stop_dfs = [[None, stopdf]]

        ssdList = []
        for lvl, df in stop_dfs:
            sdf = analyze.get_model_ssds(df, conds=self.conds, ssd_method=self.ssd_method, bwfactors=bwfactors)
            if lvl is not None:
                sdf[bwfactors] = lvl
            ssdList.append(sdf)

        self.ssdDF = pd.concat(ssdList, ignore_index=True)


    def make_headers(self, ssd_list=None):
        g_cols = self.groups
        if 'ssd' in self.data.columns:
            # get ssd's for fits if in datacols
            stopdf = self.data[self.data.ttype=='stop']
            if 'probe' in stopdf.columns and self.ssd_method=='all':
                stopdf = stopdf[stopdf.probe==1]
            self.set_model_ssds(stopdf)
            if self.ssd_method=='all':
                get_df_ssds = lambda df: np.round(df.ssd.unique(), 1).astype(int)
                ssds = [get_df_ssds(df) for _, df in stopdf.groupby(g_cols)]
                ssd_list = [np.sort(issd).tolist() for issd in ssds]
            else:
                ssd_list = [['sacc'] for i in range(self.nrows)]
        self.make_idx_cols(ssd_list)
        masterDF_header = g_cols + self.p_cols + self.q_cols
        return masterDF_header


    def make_idx_cols(self, ssd_list=None):
        """ make idx-specific headers in event of missing data
        make all other headers if not done yet
        """
        if not hasattr(self, 'q_cols'):
            self.make_q_cols()
        if not hasattr(self, 'p_cols'):
            self.make_p_cols(ssd_list)
        if not hasattr(self, 'f_cols'):
            self.make_f_cols()
        if ssd_list:
            acc = [self.p_cols[0]]
            self.idx_cols = [acc + issd + self.q_cols for issd in ssd_list]
        else:
            self.idx_cols = [self.p_cols + self.q_cols]*self.nrows


    def make_q_cols(self):
        """ make header names for correct/error RT quants
        in observedDF, yhatdf, and wtsDF
        """
        cq = ['c' + str(int(n * 100)) for n in self.model.quantiles]
        eq = ['e' + str(int(n * 100)) for n in self.model.quantiles]
        self.q_cols = cq + eq


    def make_p_cols(self, ssd_list=None):
        """ make header names for response accuracy in observedDF,
        yhatdf, and wtsDF (including SSDs if stop model)
        """
        self.p_cols = ['acc']
        if ssd_list:
            ssd_unique = np.unique(np.hstack(ssd_list)).tolist()
            self.p_cols = self.p_cols + ssd_unique


    def make_f_cols(self):
        """ make header names for various fit statistics in fitdf
        (model parameters, goodness-of-fit measures, etc)
        """
        self.poptdf_cols = np.sort(list(self.inits)).tolist()
        self.f_cols = ['idx', 'pvary', 'nvary', 'AIC', 'BIC', 'nfev', 'df', 'ndata', 'chi', 'rchi', 'logp', 'cnvrg']


    def save_results(self, saveobserved=False):
        """ Saves yhatdf and fitdf results to model output dir
        ::Arguments::
            saveobserved (bool):
                if True will write observedDF & wtsDF to
                model output dir
        """
        fname = self.model.model_id
        if self.model.is_nested:
            fname='nested_models'
        make_fname = lambda savestr: os.path.join(self.resultsdir, '_'.join([fname, savestr+'.csv']))
        yName, fName, pName = [make_fname(dfType) for dfType in ['yhat', 'finfo', 'popt']]
        self.model.yhatdf.to_csv(yName, index=False)
        self.model.fitdf.to_csv(fName, index=False)
        self.model.poptdf.to_csv(pName, index=False)
        if saveobserved:
            self.observedDF.to_csv(os.path.join(self.resultsdir, make_fname('observed_data')))
            self.wtsDF.to_csv(os.path.join(self.resultsdir, make_fname('cost_weights')))


    def make_results_dir(self, custompath=None, get_path=False):
        """ make directory for writing model output and figures
        dir is named according to model_id, navigate to dir
        after ensuring it exists
        """
        self.resultsdir = os.path.abspath(os.path.expanduser('~'))
        if custompath is not None:
            self.resultsdir = os.path.join(self.resultsdir, custompath)
        elif self.model.is_nested:
            self.resultsdir = os.path.join(self.resultsdir, "nested_models")
        else:
            self.resultsdir = os.path.join(self.resultsdir, self.model.model_id)
        if not os.path.isdir(self.resultsdir):
            os.makedirs(self.resultsdir)
        if get_path:
            return self.resultsdir



#!/usr/local/bin/env python
from __future__ import division
import numpy as np
from numpy import array
from numpy.random import sample as rs
from numpy import newaxis as na
import pandas as pd
import seaborn as sns
import matplotlib.pyplot as plt
from mpl_toolkits.axes_grid1 import make_axes_locatable
from copy import deepcopy
import matplotlib as mpl
from IPython.display import display, Latex

def plot_rt_dists(simdf, axes=None):
    targets=['A', 'B', 'C', 'D']
    targetColors = dict(zip(targets, ['#3572C6',  '#c44e52', '#8172b2', '#83a83b']))
    sns.set(style='white')
    if axes is None:
        f, axes = plt.subplots(2, 2, figsize=(9, 6), sharex=True)
    axes = axes.flatten()
    for i, ax in enumerate(axes):
        target = targets[i]
        rts = simdf[simdf.choice==target].rt.values
        sns.distplot(rts, kde=False, hist_kws={'alpha':.9}, norm_hist=True, bins=10, ax=ax, color=targetColors[target])
        top = ax.get_ylim()[1]*.75
        ax.text(750, top,  target, color=targetColors[target], fontsize=19)
    x = np.array([0,300,600,900])
    axes = np.asarray(f.axes)
    axes[0].set_ylabel('Probability Mass', fontsize=17)
    axes[2].set_ylabel('Probability Mass', fontsize=17)
    axes[2].set_xlabel('Time (ms)', fontsize=17)
    axes[3].set_xlabel('Time (ms)', fontsize=17)
    for ax in axes.flatten():
        ax.set_title('')
        ax.set_xticks(x)
        ax.set_yticklabels('')
        ax.set_xlim(0,900)
    axes[2].set_xticklabels(x, fontsize=12)
    axes[3].set_xticklabels(x, fontsize=12)
    sns.despine()

def plot_traces_rts(p, all_traces, rts, names=['A', 'B', 'C', 'D'], tb=1000):
    tr = np.mean(p['tr'])*1e3
    rtkeys = np.sort(rts.keys())
    rt_dists = [np.asarray(rts[k])*1e3-tr for k in rtkeys]
    tb = np.ceil(np.max([np.max(rti) if len(rti)>0 else 0 for rti in rt_dists]))+50
    sns.set(style='white', font_scale=1.5)
    f, axes = build_multi_axis(p, tb=tb)
    clrs = ['#3572C6',  '#c44e52', '#8172b2', '#83a83b']
    for i in range(len(all_traces)):
        for ii, ax in enumerate(axes.flatten()):
            x=np.arange(len(all_traces[i][ii]))
            ax.plot(x, all_traces[i][ii], color=clrs[ii], alpha=.3, lw=.75)

    for i, ax in enumerate(axes.flatten()):
        divider = make_axes_locatable(ax)
        axx = divider.append_axes("top", size=.7, pad=0.01, sharex=ax)
        for spine in ['top', 'left', 'bottom', 'right']:
            axx.spines[spine].set_visible(False)
        axx.set_xticklabels([])
        axx.set_yticklabels([])
        if len(rt_dists[i])<=1:
            continue
        sns.distplot(rt_dists[i], ax=axx, label=k, kde=True, hist=True, color=clrs[i], bins=20)
        text_str='$\mu_{%s}=%.fms$'%(names[i], tr+np.mean(rt_dists[i]))
        ax.text(x[0]-50, np.mean(p['a'])-.1*np.mean(p['a']), text_str, fontsize=21)

def plot_summary(outcomes, plot_traces=False, p=None, tb=1000, ntime=5):
    targets=['A', 'B', 'C', 'D']
    sns.set_palette('muted')
    f, axes = plt.subplots(3, 2, figsize=(12,14))
    a1, a2, a3, a4, a5, a6 = axes.flatten()
    choices, rts, all_traces, qdict, choicep, vdhist, vihist = outcomes
    names = np.sort(qdict.keys())
    clrs = ['#3572C6',  '#c44e52', '#8172b2', '#83a83b']
    targetColors = dict(zip(targets,clrs))
    choices = np.asarray(choices)
    counts = [len(choices[choices==i]) for i in range(4)]
    rtMeans = [np.mean(rts[target])*1000 for target in targets]
    targetTC = {target:np.array([np.mean(ts)*1000 for ts in np.array_split(np.asarray(rts[target]), ntime)]) for target in targets}

    FreqDict = {'Low':np.mean([targetTC['A'], targetTC['C']], axis=0),
               'High': np.mean([targetTC['B'], targetTC['D']], axis=0)}
    ValDict = {'Low':np.mean([targetTC['A'], targetTC['B']], axis=0),
            'High': np.mean([targetTC['C'], targetTC['D']], axis=0)}

    sns.barplot(x=targets, y=rtMeans, ax=a1, order=targets, palette=targetColors, ci=0)
    sns.barplot(x=targets, y=counts, ax=a2, order=targets, palette=targetColors, ci=0)
    x = np.arange(ntime).astype(int)
    a3.plot(x, FreqDict['Low'], label='Low', color="#3498db")
    a3.plot(x, FreqDict['High'], label='High', color="#9B59B6")
    a4.plot(x, ValDict['Low'], label='Low', color="#e74c3c")
    a4.plot(x, ValDict['High'], label='High', color="#27ae60")

    a1.set_ylabel('Response Time (ms)', fontsize=14)
    a2.set_ylabel('# Times Chosen', fontsize=14)
    a3.set_xlabel('Trial Blocks', fontsize=14)
    a4.set_xlabel('Trial Blocks', fontsize=14)
    a3.set_ylabel('RT', fontsize=14)
    a4.set_ylabel('RT', fontsize=14)
    # plt.subplots_adjust(wspace=.4)

    for i, n in enumerate(names):
        # a3.plot(np.array(qdict[n])*100, label=targets[i], color=clrs[i])
        # a4.plot(choicep[n], label=targets[i], color=clrs[i])
        a5.plot(vdhist[n], label=targets[i], color=clrs[i])
        a6.plot(vihist[n], label=targets[i], color=clrs[i])
    a3.legend(loc=0)
    a4.legend(loc=0)
    f.subplots_adjust(hspace=.35, wspace=.4)
    titles = ['$v^G_t$', '$v^N_t$']
    for i, ax in enumerate(axes.flatten()[-2:]):
        ax.set_title(titles[i])
        sns.despine(ax=ax)
    if plot_traces and p is not None:
        plot_traces_rts(p, all_traces, rts, tb=tb)
    sns.despine()

def plot_simdf_summary(simdf):
    f, axes = plt.subplots(2, 2, figsize=(12,8))
    a1, a2, a3, a4 = axes.flatten()
    targets=['A', 'B', 'C', 'D']
    clrs = ['#3572C6',  '#c44e52', '#8172b2', '#83a83b']
    targetColors = dict(zip(targets,clrs))
    sns.barplot(x='choice', y='rt', data=simdf, ax=a1, order=targets, palette=targetColors)
    sns.barplot(x='choice', y='switch', data=simdf, ax=a2, order=targets, palette=targetColors)
    a1.set_ylabel('Response Time (ms)', fontsize=13)
    a2.set_ylabel('P(Switch)', fontsize=13)
    rts = simdf.groupby('choice').mean().rt.values
    sw = simdf.groupby('choice').mean().switch.values
    a1.set_ylim(rts.min()*.85, rts.max()*1.15)
    a2.set_ylim(sw.min()*.50, sw.max()*1.20)
    for i, target in enumerate(targets):
        tcolor=targetColors[target]
        tdf = simdf[simdf.choice=='target'].reset_index()
        sns.timeseries.tsplot(data=simdf, time='trial', unit='agent', value='vd'+target, ax=a3, color=tcolor)
        sns.timeseries.tsplot(data=simdf, time='trial', unit='agent', value='vi'+target, ax=a4, color=tcolor)
    a3.legend(loc=0)
    f.subplots_adjust(hspace=.35, wspace=.4)
    a3.set_ylabel('$v^G_t$', fontsize=16)
    a4.set_ylabel('$v^N_t$', fontsize=16)
    a3.set_xlabel('Trial ( $t$ )', fontsize=13)
    a4.set_xlabel('Trial ( $t$ )', fontsize=13)
    plt.subplots_adjust(wspace=.4)
    sns.despine()

def get_avg_slope_trace(traces, nalt=4):
    slope_func = lambda x0,x1,y0,y1: (y1-y0)/(x1-x0)
    rise, run, slopes=[], [], []
    for alt_n in range(nalt):
        y0 = 0
        y1 = np.mean([traces_i[alt_n][-1] for traces_i in traces])
        x0 = 0
        x1 = np.ceil(np.mean([traces_i[alt_n].shape[0] for traces_i in traces]))
        rise.append(y1)
        run.append(x1)
        slopes.append(slope_func(x0,x1,y0,y1))

    return rise, run, slopes


def gen_mappable(vals_to_map, cm='rainbow'):

    ncolors = len(vals_to_map)
    clrs = sns.color_palette(cm, ncolors)
    mycmap = mpl.colors.ListedColormap(clrs)

    vmin = np.min(vals_to_map)
    vmax = np.max(vals_to_map)
    norm = mpl.colors.Normalize(vmin=vmin, vmax=vmax)

    scalar_mappable = mpl.cm.ScalarMappable(cmap=mycmap, norm=norm)
    scalar_mappable.set_array([])

    return scalar_mappable, vmin, vmax


def plot_reactivity_strategy(trialsdf, igtdf, cm='rainbow', save=False, pq='P', plot_scatter=False):

    n = trialsdf.a_go.unique().size
    a_go = np.sort(trialsdf.a_go.unique())
    sm, vmin, vmax = gen_mappable(vals_to_map=a_go, cm=cm)
    if plot_scatter:
        f, axes = plt.subplots(2,2, figsize=(12, 9))
        ax1, ax2, ax3, ax4 = axes.flatten()
    else:
        f, axes = plt.subplots(1, 2, figsize=(12, 4.5))
        ax1, ax2 = axes.flatten()

    for grp, grpdf in trialsdf.groupby('a_go'):
        colr = sm.to_rgba(grp)
        ax1.plot(grpdf.vdiff.values, color=colr)
        ax2.plot(grpdf.v_opt_diff.values, color=colr)
        sns.despine()

    divider = make_axes_locatable(ax2)
    cax = divider.append_axes("right", size="5%", pad=0.1)
    cb = plt.colorbar(sm, cax)
    sm.colorbar.set_ticks([vmin, vmax])
    cax.set_yticklabels(['{:.3f}'.format(vmin), '{:.3f}'.format(vmax)])

    ax1.set_title('$\Delta_{card} = V_{D}(t)-V_{I}(t)$', fontsize=15)
    ax2.set_title('$\Delta_{OS} = (\Delta_C + \Delta_D) - (\Delta_A + \Delta_B)$', fontsize=14)
    ax1.set_ylabel('Selected Channel Strength ($\Delta_{card}$)', fontsize=12)
    ax2.set_ylabel('Optimal-Subopt Ch. Strength ($\Delta_{OS}$)', fontsize=12)

    for ax in [ax1, ax2]:
        ax.set_xlabel('Trials', fontsize=15)
    if plot_scatter:
        if pq=='P':
            measure = 'Payoff'
        else:
            measure = 'Sensitivity'
        pvals_by_group = igtdf.groupby('a_go').mean().loc[:, pq].values
        reactivity = np.array([grpdf.vdiff.mean() for grp, grpdf in trialsdf.groupby('a_go')])
        strategy = np.array([grpdf.v_opt_diff.mean() for grp, grpdf in trialsdf.groupby('a_go')])
        for i in range(n):
            colr = sm.to_rgba(a_go[i])
            ax3.scatter(pvals_by_group[i], reactivity[i], color=colr, s=15)
            ax4.scatter(pvals_by_group[i], strategy[i], color=colr, s=15)
        ax3.set_ylabel('$\mu \Delta_{card}$', fontsize=15)
        ax4.set_ylabel('$\mu \Delta_{OS}$', fontsize=15)
        ax3.set_xlabel(measure, fontsize=15)
        ax4.set_xlabel(measure, fontsize=15)

    f.subplots_adjust(wspace=.3, hspace=.2)
    plt.tight_layout()
    if save:
        savestr = "_aN" + str(trialsdf.a_no.unique()[0])
        f.savefig(''.join(['reactivity_strategy_', measure, savestr, '.png']), dpi=400)


def build_multi_axis(p, nresp=4, tb=800):
    sns.set(style='white', font_scale=1.5)
    bound = p['a']
    onset = p['tr']
    if hasattr(bound, '__iter__'):
        bound = bound[0]
        onset = onset[0]
    # init figure, axes, properties
    f, axes = plt.subplots(2, 2, figsize=(14, 7), sharex=True, sharey=True, dpi=600)
    f.subplots_adjust(hspace=.1, top=.99, bottom=.05)
    w = tb + 40
    h = bound
    start = onset - 80
    axes=axes.flatten()
    # c=["#e74c3c", '#27ae60', '#4168B7', '#8E44AD']

    for i, ax in enumerate(axes):
        plt.setp(ax, xlim=(start - 1, w + 1), ylim=(0 - (.01 * h), h + (.01 * h)))
        ax.hlines(y=h, xmin=start, xmax=w, color='k')
        ax.hlines(y=0, xmin=start, xmax=w, color='k')
        ax.vlines(x=tb, ymin=0, ymax=h, color='#2043B0', lw=1, linestyle='-', alpha=.35)
        ax.vlines(x=start + 2, ymin=0, ymax=h, color='k')
        ax.set_xticklabels([])
        ax.set_yticklabels([])
        ax.set_xticks([])
        ax.set_yticks([])
    sns.despine(top=True, right=True, bottom=True, left=True)
    return f, axes



#!/usr/local/bin/env python
# Authors:
#   Kyle Dunovan (ked64@pitt.edu) &
#   Jeremy Huang (jeremyhuang@cmu.edu)

from __future__ import division
from future.utils import listvalues
import os, sys
import numpy as np
import pandas as pd
from radd import models, theta
from radd.adapt import models_rl
from radd.tools import messages, utils
from copy import deepcopy
from scipy.optimize import basinhopping, differential_evolution, fmin
from numpy.random import uniform
from lmfit import minimize, fit_report
from IPython.display import clear_output


class GlobalBounds(object):
    """ sets conditions for step acceptance during
    basinhopping optimization routine
    Arguments:
        xmin (list): lower boundaries for each parameter
        xmax (list): upper boundaries for each parameter
    """

    def __init__(self, xmin, xmax):
        self.xmin = np.array(xmin)
        self.xmax = np.array(xmax)

    def __call__(self, **kwargs):
        x = kwargs["x_new"]
        tmin = bool(np.all(x >= self.xmin))
        tmax = bool(np.all(x <= self.xmax))
        return tmin and tmax


class HopStep(object):
    """ scale stepsize of basinhopping optimization according to
    individual parameters ranges (smaller steps for more sensitive params).
    See theta.get_stepsize_scalars() for parameter <--> scalar mapping.
    Arguments:
        keys (list): list of parameter names
        nlevels (list): list of levels per parameter
        stepsize (list): initial stepsize
    """

    def __init__(self, keys, nlevels, stepsize=.15):
        self.stepsize_scalars = self.get_stepsize_scalars(keys, nlevels)
        self.stepsize = stepsize
        self.np = self.stepsize_scalars.size
        self.stepsizeList = []

    def get_stepsize_scalars(self, keys, nlevels):
        """ returns an array of scalars used by for controlling
        stepsize of basinhopping algorithm for each parameter
        """
        scalar_dict={'a': .075,
                    'tr': .025,
                    'v': .18,
                    'ssv': .2,
                    'xb': .3,
                    'sso': .01,
                    'z': .05,
                    'si': .005,
                    'C': .002,
                    'B': .008,
                    'R': .002,
                    'vi': 1.1,
                    'vd': 1.5,
                    'Beta': .1}

        nl = [np.ones(nl) for nl in nlevels]
        stepsize_scalars = np.hstack([scalar_dict[k]*nl[i] for i,k in enumerate(keys)])
        stepsize_scalars = stepsize_scalars.squeeze()
        return stepsize_scalars

    def __call__(self, x):
        s = self.stepsize
        self.stepsizeList.append(s)
        ss = self.stepsize_scalars
        x = np.array([x[i] + uniform(-ss[i]*s, ss[i]*s) for i in range(self.np)])
        return x


def format_local_bounds(xmin, xmax):
    """ groups (xmin, xmax) for each parameter """
    tupler = lambda xlim: tuple([xlim[0], xlim[1]])
    # return map((tupler), zip(xmin, xmax))
    return [tupler(xl) for xl in zip(xmin, xmax)]


def format_basinhopping_bounds(basin_keys, nvary, kind='xdpm', tb=None):
    """ returns separate lists of all parameter
    min and max values """
    allbounds = theta.get_bounds(kind=kind, tb=tb)
    xmin, xmax = [], []
    nlevels = [np.ones(nl) for nl in nvary]
    for i, pk in enumerate(basin_keys):
        xmin.append([allbounds[pk][0]] * nlevels[i])
        xmax.append([allbounds[pk][1]] * nlevels[i])
    xmin = np.hstack(xmin).tolist()
    xmax = np.hstack(xmax).tolist()
    return xmin, xmax



class Optimizer(object):
    """ Primary optimization class for handling global and local optimization routines.

    Arguments:
        fitparams (Series): fitparams Series object
        inits (dict): parameter dictionary to initialize model with
        param_sets (list): list of init dicts to run basinhopping over
        basinparams (dict): dictionary of global optimization params
        progress (bool): initialize progress bars (default=True)
        custompath (str): local path from ~ to save results
        custompath (str): local path from ~ to save results
    """

    def __init__(self, fitparams={}, basinparams={}, inits=None, param_sets=None, custompath=None, nruns=10, data=None):

        self.fitparams = fitparams
        self.basinparams = basinparams
        self.kind = fitparams.kind
        self.learn = fitparams.learn
        self.depends_on = fitparams.depends_on
        self.ssdMethod = fitparams.ssd_method
        self.data = data

        if inits is None:
            inits = theta.get_default_inits(kind=self.kind, depends_on=self.depends_on, learn=self.learn)

        self.inits = deepcopy(inits)
        self.callback = None
        self.constants = list(inits)
        self.param_sets = param_sets
        self.nruns=nruns

        self.progress = self.basinparams['progress']
        if self.progress:
            self.make_progress_bars(inits=True, basin=True)

        if self.learn:
            data = self.get_trials_data()
            self.simRL = models_rl.Simulator(self.inits, data=data, fitparams=self.fitparams, ssdMethod=self.ssdMethod, constants=self.constants)

        self.sim = models.Simulator(self.inits, fitparams=self.fitparams, ssdMethod=self.ssdMethod)
        self.update()
        self.make_results_dir(custompath=custompath)


    def update(self, ksData=None, **kwargs):
        kw_keys = list(kwargs)
        kw = pd.Series(kwargs)
        if 'fitparams' in kw_keys:
            self.fitparams = kw.fitparams
        if 'basinparams' in kw_keys:
            self.basinparams = kw.basinparams

        self.ix = self.fitparams['ix']
        self.nlevels = self.fitparams['nlevels']
        self.learn = self.fitparams['learn']
        self.pcmap = self.fitparams.pcmap
        self.inits = self.fitparams.inits
        self.progress = self.basinparams['progress']
        self.ksTest = False

        if ksData is not None:
            self.ksTest = True

        if 'nruns' in kw_keys:
            self.nruns = kw.nruns

        if self.learn:
            constants = self.simRL.constants
            if 'constants' in kw_keys:
                constants = kw['constants']
            data = self.get_trials_data()
            self.simRL.update(inits=self.inits, fitparams=self.fitparams, data=data, constants=constants, nruns=self.nruns)

        self.sim.update(inits=self.inits, fitparams=self.fitparams, ksData=ksData)


    def hop_around(self, param_sets=None, learn=False, ratesOnly=True, fitDynamics=True, rateParams=['B', 'C', 'R']):
        """ initialize model with niter randomly generated parameter sets
        and perform global minimization using basinhopping algorithm
        ::Arguments::
            p (dict):
                parameter dictionary
        ::Returns::
            parameter set with the best fit
        """
        if param_sets is None:
            self.sample_param_sets(learn)
            param_sets = self.param_sets
        if self.progress:
            self.make_progress_bars(inits=True, basin=True)

        xpopt, xfmin, global_results = [], [], []
        self.gpopt = []
        for i, p in enumerate(param_sets):
            self.update(inits=p, force='flat', learn=learn)

            if self.progress:
                self.ibar.update(value=i, status=i+1)
            popt, fmin, out = self.optimize_global(p=p, learn=learn, ratesOnly=ratesOnly, fitDynamics=fitDynamics, rateParams=rateParams, resetProgress=False, return_all=True)
            xpopt.append(popt)
            self.gpopt.append(deepcopy(popt))
            xfmin.append(fmin)
            global_results.append(out)

        if self.progress:
            self.gbar.clear()
            self.ibar.clear()

        keep_ix = np.argmin(xfmin)
        popt = xpopt[keep_ix]
        self.global_popt = deepcopy(popt)
        self.global_fmin = xfmin[keep_ix]
        self.global_results = global_results[keep_ix]

        if learn:
            self.rlpopt = deepcopy(popt)
        else:
            self.popt = deepcopy(popt)
        return popt


    def popt_array_to_dict(self, popt_arr, learn=False):
        if learn:
            pdict = dict(zip(self.simRL.pvary, popt_arr))
            pdict.update(self.simRL.fixedParams)
        elif self.sim.nlevels>1:
            nvary = self.sim.nvary
            pvary = self.sim.pvary
            pdict = {}; start = 0
            for i, n in enumerate(nvary):
                pdict[pvary[i]] = np.array(popt_arr[start:start+n])
                start = n
        else:
            pdict = dict(zip(self.sim.pvary, popt_arr))
        return pdict


    def optimize_global(self, p, learn=False, fitDynamics=True, ratesOnly=True, rateParams=['B', 'C', 'R'], resetProgress=False, return_all=False):
        """ Global optimization with basinhopping (or differential_evolution)
        ::Arguments::
            p (dict):               parameter dictionary
            learn (bool):           fit adaptive model if True (default: False)
            fitDynamics (bool):     fit adaptive model to timeseries of rt/accuracy
            ratesOnly (bool):       restrict free parameters to learning rates
            rate params (list):     learning rate parameters
            resetProgress (bool):   reset global progress bar
            return_all (bool):      return all results from optimzation
        ::Returns::
            popt (dict):            optimized parameter dictionary
            fmin (float):           function minimum
        """
        bp = self.basinparams
        popt = deepcopy(p)
        if learn:
            if fitDynamics:
                costfx = self.simRL.cost_fx_rl
            else:
                costfx = self.simRL.cost_fx
            if ratesOnly:
                pkeys = [k for k in list(p) if k not in rateParams]
            else:
                pkeys = []
            self.simRL.update(inits=p, constants=pkeys)
            x0 = self.simRL.preproc_params(p, asarray=True)
        else:
            self.sim.set_pconstant_values_matrix(p)
            x0 = self.sim.pdict_to_array(p)
            costfx = self.sim.cost_fx
            if self.ksTest:
                costfx = self.sim.ks_stat

        if self.progress:
            if resetProgress:
                # print("")
                self.make_progress_bars(inits=True, basin=True)
                clear_output()
            self.callback = self.gbar.reset(get_call=True, gbasin=resetProgress)

        # create args for customizing global optimizer
        self.set_global_options(learn=learn)

        # run global optimization (basinhopping/differential_evolution)
        if bp['method']=='basin':
            out = basinhopping(costfx, x0=x0, minimizer_kwargs=self.polish_args, T=bp['T'], stepsize=bp['stepsize'], niter_success=bp['nsuccess'], niter=bp['niter'], interval=bp['interval'], take_step=self.custom_step, accept_test=self.accept_step, callback=self.callback)
            fit_info = out.lowest_optimization_result

        elif bp['method']=='evolution':
            out = differential_evolution(costfx, bounds=self.polish_args['bounds'], popsize=bp['popsize'], recombination=bp['recombination'], mutation=bp['mutation'], strategy=bp['strategy'], disp=bp['disp'], polish=True, maxiter=bp['maxiter'], tol=bp['tol'], callback=self.callback, atol=self.fitparams['tol'])
            if self.progress:
                self.gbar.clear()
            fit_info = out

        pdict = self.popt_array_to_dict(fit_info.x,  learn=learn)
        popt.update(pdict)
        fmin = fit_info.fun
        if return_all:
            return popt, fmin, out

        return popt, fmin


    def gradient_descent(self, p, learn=False, flat=False):
        """ Local optimization with Nelder-Mead Simplex algorithm (gradient descent)
        ::Arguments::
            p (dict):       parameter dictionary
            learn (bool):   fit adaptive model if True (default: False)
            flat
        ::Returns::
            finfo (Series): summary stats of optimization
            popt (dict):    optimized parameter dictionary
            yhat (array):   model-predicted data vector
        """
        flat= False

        fp = self.fitparams
        fp['inits'] = p
        self.update(fitparams=fp, ksData=self.sim.ksData)
        fp = self.fitparams
        if fp.nlevels==1:
            flat=True

        optkws = {'xtol': fp['tol'], 'ftol': fp['tol'], 'maxfev': fp['maxfev'], 'maxiter': fp['maxiter']}
        if learn:
            sim = self.simRL
            pflat = sim.pflat
            pvary = sim.pvary
            x0 = sim.preproc_params(p, asarray=True)
            lmParams = theta.loadParameters_RL(inits=p, pflat=pflat, pvary=pvary, kind=self.kind)
            costfx = sim.cost_fx_lmfit
        else:
            sim = self.sim
            sim.set_pconstant_values_matrix(p)
            lmParams = theta.loadParameters(inits=p, pcmap=self.pcmap, is_flat=flat, kind=self.kind)
            lmParamsNames = list(lmParams.valuesdict())
            sim.update(lmParamsNames=lmParamsNames, ksData=sim.ksData)
            costfx = sim.cost_fx_lmfit
            # if self.ksTest:
                # costfx = sim.ks_stat_lmfit

        self.lmParams = lmParams

        if self.progress:
            self.lbar = utils.GradientCallback(n=fp['maxfev'], fmin=10000)
            self.lcallback = self.lbar.callback
        else:
            self.lcallback = None

        if 'least' in fp['method']:
            self.lmMin = minimize(costfx, lmParams, method=fp['method'], iter_cb=self.lbar.callback)
        elif fp['method']=='brute':
            #rranges = (slice(-4, 4, 0.25), slice(-4, 4, 0.25))
            self.lmMin = minimize(costfx, lmParams, method=fp['method'], iter_cb=self.lbar.callback, Ns=20)
        else:
            self.lmMin = minimize(costfx, lmParams, method=fp['method'], tol=fp['tol'],  options=optkws, iter_cb=self.lbar.callback)
        if hasattr(self, 'lbar'):
            self.lbar.clear()

        self.param_report = fit_report(self.lmMin.params)
        return self.assess_fit(flat=flat, learn=learn)


    def set_global_options(self, learn=False):
        if learn:
            sim = self.simRL
        else:
            sim = self.sim
        bp = self.basinparams
        fp = self.fitparams
        xmin, xmax = format_basinhopping_bounds(sim.pvary, sim.nvary, self.kind, tb=fp['tb'])
        bounds = format_local_bounds(xmin, xmax)

        # define custom take_step and accept_test functions
        self.accept_step = GlobalBounds(xmin, xmax)
        self.custom_step = HopStep(sim.pvary, nlevels=sim.nvary, stepsize=bp['stepsize'])
        self.polish_args = {"method": bp['local_method'], 'bounds': bounds, 'tol': bp['polish_tol'], 'options': {'xtol': bp['polish_tol'], 'ftol': bp['polish_tol']}}

        if self.progress:
            if not hasattr(self, 'gbar') or self.gbar.pbar.displayed==False:
                self.make_progress_bars(inits=False, basin=True)
            self.callback = self.gbar.reset(get_call=True, gbasin=False, history=False)


    def sample_param_sets(self, learn=False, fitDynamics=True):
        if learn:
            self.sample_rl_theta(fitDynamics=fitDynamics)
        else:
            self.sample_theta()


    def sample_theta(self):
        """ sample *nsamples* (default=5000, see set_fitparams) different
        parameter sets (param_sets) and get model yhat for each set (param_yhats)
        """
        pkeys = self.sim.theta.keys().tolist()
        nsamples = self.basinparams['nsamples']
        if not hasattr(self, 'init_params'):
            init_params = theta.random_inits(pkeys, ninits=nsamples, kind=self.kind, as_list=True, method=self.basinparams['sample_method'])
            init_yhats = pd.DataFrame(np.vstack([self.sim.simulate_model(p) for p in init_params]))
            self.init_params = init_params
            self.init_yhats = init_yhats.copy()
        init_params = self.init_params
        init_yhats = self.init_yhats.copy()
        psets = self.filter_params(init_params, init_yhats)
        self.param_sets = psets


    def sample_rl_theta(self, fitDynamics=True):
        if fitDynamics:
            costfx = self.simRL.cost_fx_rl
        else:
            costfx = self.simRL.cost_fx
        nsamples = self.basinparams['nsamples']
        nkeep = self.basinparams['ninits']
        idxParams = pd.DataFrame([pd.Series(self.popt)]*nsamples)
        rlParams = theta.random_inits(['B', 'C', 'R'], ninits=nsamples, kind=self.kind, as_list=True, method=self.basinparams['sample_method'])
        rlParamsDF = pd.concat([idxParams, pd.DataFrame(rlParams.tolist())], axis=1)
        rlParams = np.array(rlParamsDF.to_dict('records'))

        self.simRL.update(inits=self.popt, constants=list(self.popt))
        cost = np.vstack([costfx(p) for p in rlParams])
        rlPsets = rlParams[np.argsort(np.hstack(cost))[:nkeep]]
        self.param_sets = rlPsets


    def filter_params(self, psets=None, yhatDF=None):
        """ sample *nsamples* (default=5000, see set_fitparams) different
        parameter sets (param_sets) and get model yhat for each set (param_yhats)
            if fit_on==subjects flat_y shape is (n_idx X ndata)
            elseif fit_on==average flat_y shape is (1 X ndata)
        """

        if psets is None:

            psets = self.init_params
            yhatDF = self.init_yhats.copy()

        nkeep = self.basinparams['ninits']
        y = self.fitparams.y
        wts = self.fitparams.wts
        keys = yhatDF.columns.tolist()

        ySeries = pd.Series(dict(zip(keys, y)))
        wSeries = pd.Series(dict(zip(keys, wts)))
        psets = np.asarray(psets)
        diff = yhatDF - ySeries
        wDiff = diff * wSeries

        sqDiff = wDiff.apply(lambda x: x**2, axis=0)
        sseDF = sqDiff.apply(lambda x: np.sum(x), axis=1)
        sseVals =sseDF.values
        bestIX = sseVals.argsort()[:nkeep]

        return psets[bestIX]


    def get_trials_data(self):
        if self.fitparams.fit_on=='subjects':
            idx = np.int(self.fitparams['idx'])
            return self.data[self.data.idx==idx]
        else:
            return self.data.copy()


    def assess_fit(self, flat=True, learn=False):
        """ wrapper for analyze.assess_fit calculates and stores
        rchi, AIC, BIC and other fit statistics
        ::Arguments::
            flat (bool):
                if flat, yhat have ndim=1, else ndim>1
        ::Returns::
            yhat (array), finfo (pd.Series), popt (dict)
            see gradient_descent() docstrings
        """
        fp = deepcopy(self.fitparams)
        if learn:
            y = np.hstack([self.simRL.rtBlocks, self.simRL.saccBlocks])
            wts = np.hstack([self.simRL.rt_weights, self.simRL.sacc_weights])
            sim = self.simRL
        else:
            y = self.sim.y.flatten()
            wts = self.sim.wts.flatten()
            sim = self.sim

        # gen dict of lmfit optimized Parameters object
        popt = dict(self.lmMin.params.valuesdict())
        fmin = self.lmMin.chisqr
        nvary = len(self.lmMin.var_names)
        residuals = self.lmMin.residual
        yhat = (residuals / wts) + y
        if self.lmMin.method=='brute':
            success = not self.lmMin.aborted
        else:
            success = self.lmMin.success
        nfev = self.lmMin.nfev
        try: niter = self.lmMin.nit
        except Exception: niter = nfev
        # resContainer = self.lmMin
        #
        # # check if global optimization better
        # if hasattr(self, "global_popt"):
        #     checkGlobal = list(self.global_popt) == list(popt)
        #     if fmin > self.global_fmin and checkGlobal:
        #         popt = deepcopy(self.global_popt)
        #         fmin = self.global_fmin
        #         resContainer = self.global_results.lowest_optimization_result
        #         resContainer['nfev'] = self.global_results.nfev
        #         resContainer['nit'] = self.global_results.nit
        #
        # residualList = []
        # for i in range(5):
        #     # sim.update(inits=popt)
        #     yhat = sim.simulate_model(popt)
        #     residualList.append(wts * (yhat - y))
        #
        # residual = np.mean(residualList, axis=0)
        # yhat = (residual / wts) + y
        # success = resContainer.success
        # nfev = resContainer.nfev
        # niter = resContainer.nit

        # TODO: extract, calculate, and store std.errors of popt
        # presults is scipy.minimize object (see hop_around() for global_results)
        # presults = self.global_results.lowest_optimization_result
        # then take sqrt of the diag. of the hessian to get errors
        # poptErr = np.sqrt(np.diag(presults.hess_inv.todense()))

        if not self.learn:
            # un-vectorize all parameters except conditionals
            popt = theta.scalarize_params(popt, self.pcmap)
            if fp.nlevels>1:
                popt = theta.pvary_levels_to_array(popt, self.pcmap)

        finfo = pd.Series()
        # get model-predicted yhat vector
        fp['yhat'] = yhat
        finfo['idx'] = fp.idx
        finfo['pvary'] = '_'.join(list(fp.depends_on))
        finfo['cnvrg'] = success
        finfo['nfev'] = nfev
        finfo['niter'] = niter
        finfo['nvary'] = nvary
        finfo['chi'] = fmin
        finfo['ndata'] = y.size
        finfo['df'] = finfo.ndata - finfo.nvary
        finfo['rchi'] = finfo.chi / finfo.df
        finfo['logp'] = finfo.ndata * np.log(finfo.chi / finfo.ndata)
        finfo['AIC'] = finfo.logp + 2 * finfo.nvary
        finfo['BIC'] = finfo.logp + finfo.nvary * np.log(finfo.ndata)
        return finfo, popt, fp['yhat']


    def write_results(self, finfo, popt, yhat):
        """ logs fit info to txtfile, fills yhatdf and fitdf
        """
        self.yhatdf = self.model.handler.fill_yhatdf(yhat=yhat, fitparams=self.fitparams)
        self.fitdf = self.model.handler.fill_fitdf(finfo=finfo, fitparams=self.fitparams)
        self.poptdf = self.model.handler.fill_poptdf(popt=popt, fitparams=self.fitparams)


    def save_results(self, custompath=None, saveobserved=False):
        """ Saves yhatdf and fitdf results to model output dir
        ::Arguments::
            saveobserved (bool):
                if True will write observedDF & wtsDF to
                model output dir
        """
        if custompath is not None:
            self.make_results_dir(custompath=custompath)
        fname = self.model.model_id
        if self.model.is_nested:
            fname='nested_models'
        resultsPath = os.path.join(self.resultsdir, fname)
        make_fname = lambda savestr: '_'.join([resultsPath, savestr+'.csv'])
        self.yhatdf.to_csv(make_fname('yhat'), index=False)
        self.fitdf.to_csv(make_fname('finfo'), index=False)
        self.poptdf.to_csv(make_fname('popt'), index=False)
        if saveobserved:
            self.model.observedDF.to_csv(make_fname('observed_data'))
            self.model.wtsDF.to_csv(make_fname('cost_weights'))


    def log_fit_info(self, finfo=None, popt=None, yhat=None):
        """ write meta-information about latest fit
        to logfile (.txt) in working directory
        """
        fp = deepcopy(self.fitparams)
        fp['yhat'] = yhat
        # lmfit-structured fit_report to write in log file
        param_report = self.param_report
        # log all fit and meta information in working directory
        messages.logger(param_report, finfo=finfo, popt=popt, fitparams=fp, kind=fp.kind, fit_on=fp.fit_on, savepath=self.resultsdir)


    def make_results_dir(self, custompath=None):
        savedir = os.path.expanduser('~')
        if custompath is not None:
            savedir = os.path.join(savedir, custompath)
            if not os.path.isdir(savedir):
                os.mkdir(savedir)
        else:
            if 'Dropbox' in os.listdir(savedir):
                savedir = os.path.join(savedir, 'Dropbox')
        self.resultsdir = savedir


    def make_progress_bars(self, inits=True, basin=True, lBasin=True):
        bp = self.basinparams
        if inits:
            ninits = bp['ninits']
            status=' / '.join(['Inits {}', '{}'.format(ninits)])
            self.ibar = utils.PBinJ(n=ninits, color='b', status=status)
        if basin:
            if hasattr(self, 'gbar'):
                self.gbar.clear()
            if bp['method']=='basin':
                niter = bp['nsuccess']
            else:
                niter = bp['maxiter']
            self.gbar = utils.GlobalCallback(n=niter, fmin=10000, method=bp['method'])
            self.callback = self.gbar.reset(get_call=True)
        if lBasin:
            fp = self.fitparams
            if hasattr(self, 'lbar'):
                self.lbar.clear()
            self.lbar = utils.GradientCallback(n=fp['maxfev'], fmin=10000)
            self.callback = self.lbar.reset(get_call=True)



#!/usr/local/bin/env python
# Authors:
#   Kyle Dunovan (ked64@pitt.edu) &
#   Jeremy Huang (jeremyhuang@cmu.edu)
from __future__ import division
from future.utils import listvalues, listitems
from copy import deepcopy
import pandas as pd
import numpy as np
from numpy.random import random_sample as randsample
from scipy.stats import ks_2samp as KS
from scipy.stats.mstats import mquantiles
from itertools import product
from radd import theta
from radd.tools.utils import pandaify_results
from numpy import hstack as hs
from radd.compiled.jitfx import *


class Simulator(object):

    def __init__(self, inits, fitparams=None, ssdMethod='all', **kwargs):
        self.ssdMethod = ssdMethod
        self.update(fitparams=fitparams, inits=inits)
        self.ksData=None


    def update(self, force=False, ksData=None, **kwargs):
        kw_keys = list(kwargs)
        if 'fitparams' in kw_keys:
            self.fitparams = kwargs['fitparams']
        if 'inits' in kw_keys:
            self.inits = deepcopy(kwargs['inits'])
            self.theta = pd.Series(self.inits)
        if 'constants' in kw_keys:
            self.constants = deepcopy(kwargs['constants'])
        if 'lmParamsNames' in kw_keys:
            lmParamsNames = kwargs['lmParamsNames']
        else:
            lmParamsNames = None
        if ksData is not None:
            self.ksData = ksData
            nl = self.fitparams['nlevels']
            corRT = self.ksData['corRT']
            errRT = self.ksData['errRT']
            self.RT = [np.concatenate([-errRT[i], corRT[i]]) for i in range(nl)]
            self.goAcc = self.ksData['goAcc']
            self.stopAcc = self.ksData['stopAcc']

        self.kind = self.fitparams.kind
        self.clmap = self.fitparams.clmap
        self.depends_on = self.fitparams.depends_on
        self.conds = list(self.clmap)
        self.nconds = len(self.conds)
        self.quantiles = self.fitparams.quantiles
        self.tb = self.fitparams['tb']
        self.dt = self.fitparams['dt']
        self.si = self.fitparams['si']
        self.nlevels = self.fitparams['nlevels']
        self.ntrials = self.fitparams['ntrials']
        self.y = self.fitparams.y.flatten()
        self.wts = self.fitparams.wts.flatten()

        try:
            self.ssd, nssd, nss, nss_per, ssd_ix = self.fitparams.ssd_info
            nfree = self.pmatrix_vals.shape[0]
            if self.ssd.shape[0]!=nfree:
                ntile = nfree/self.ssd.shape[0]
                self.ssd = np.tile(self.ssd, ntile).reshape(nfree, -1)
        except Exception:
            pass

        self.format_cond_params(lmParamsNames=lmParamsNames)
        self.make_io_vectors()
        self.__init_analyze_functions__()



    def __init_analyze_functions__(self):
        """ initiates the analysis function used in
        optimization routine to produce the yhat vector
        """
        prob = self.quantiles
        # self.RTQ = lambda zpd: map((lambda x: mquantiles(x[0][x[0] < x[1]], prob)), zpd)
        self.RTQ = lambda zpd: [mquantiles(rt[rt < deadline], prob) for rt, deadline in zpd]
        self.filterRT = lambda zpd: [rt[rt < deadline] for rt, deadline in zpd]

    def cost_fx(self, theta_array):
        yhat = self.simulate_model(theta_array)
        return np.sum((self.wts * (yhat - self.y))**2)


    def cost_fx_lmfit(self, lmParams, sse=False):
        thetaSeries = pd.Series(lmParams.valuesdict())[self.lmParamsNames]
        yhat = self.simulate_model(thetaSeries.values)
        residuals = self.wts * (yhat - self.y)
        if sse:
            return np.sum(residuals**2)
        return residuals

    def chi_square(self, theta_array):
        rt, ssrt = self.simulate_model(theta_array, analyze=False, get_rts=True)
        ntotal = rt[0].shape[0]
        nbins = self.quantiles.size+1
        eProp= [[rt[(rt >= rtq[i])&(rt < rtq[i+1])].shape[0] / ntotal for i in range(nbins)] for ii in range(self.nconds)]
        pResp = np.mean(np.where(rt < m.sim.tb, 1, 0), axis=1)
        E_i = N * pResp * np.asarray(eProp)
        chi2 = np.sum((O_i - E_i)**2 / E_i)


    def ks_stat_lmfit(self, lmParams):
        thetaSeries = pd.Series(lmParams.valuesdict())[self.lmParamsNames]
        crtHat, ssrt = self.simulate_model(thetaSeries.values, analyze=False, get_rts=True)
        nl, nssd, nssPer = ssrt.shape
        nss = nssd * nssPer
        ertHat = crtHat[:, :nss].reshape(ssrt.shape)
        gacc = np.mean(ufunc_where(crtHat < self.tb, 1, 0), axis=1)
        sacc = np.mean(ufunc_where(ssrt <= ertHat, 1, 0), axis=2)
        crtHat = self.filterRT(zip(crtHat, [self.tb]*nl))
        ertHat = self.filterRT(zip(ertHat, ssrt))
        rtHat = [np.concatenate([-ertHat[i], crtHat[i]]) for i in range(nl)]
        ksRT = np.sum([np.log(KS(self.RT[i], rtHat[i])[0]) for i in range(nl)])
        llGo = np.sum([-np.log(np.exp(-((self.goAcc[i]-gacc[i])/.1)**2)) for i in range(nl)])
        llStop = np.sum([[-np.log(np.exp(-((self.stopAcc[i][ii]-sacc[i][ii])/.1)**2)) for ii in range(nssd)] for i in range(nl)])
        LL = ksRT + llGo + llStop
        return LL


    def ks_stat(self, theta_array):
        crtHat, ssrt = self.simulate_model(theta_array, analyze=False, get_rts=True)
        nl, nssd, nssPer = ssrt.shape
        nss = nssd * nssPer
        ertHat = crtHat[:, :nss].reshape(ssrt.shape)
        gacc = np.mean(ufunc_where(crtHat < self.tb, 1, 0), axis=1)
        sacc = np.mean(ufunc_where(ssrt <= ertHat, 1, 0), axis=2)

        crtHat = self.filterRT(zip(crtHat, [self.tb]*nl))
        ertHat = self.filterRT(zip(ertHat, ssrt))
        rtHat = [np.concatenate([-ertHat[i], crtHat[i]]) for i in range(nl)]
        ksRT = np.sum([np.log(KS(self.RT[i], rtHat[i])[0]) for i in range(nl)])
        llGo = np.sum([-np.log(np.exp(-((self.goAcc[i]-gacc[i])/.1)**2)) for i in range(nl)])
        llStop = np.sum([[-np.log(np.exp(-((self.stopAcc[i][ii]-sacc[i][ii])/.1)**2)) for ii in range(nssd)] for i in range(nl)])
        LL = ksRT + llGo + llStop
        return LL


    def simulate_model(self, params, analyze=True, get_rts=False):
        xtb, drift, ssdrift, bound, gbase, gOnset, ssOnset, dx = self.params_to_array(params, preprocess=True)
        dvg, goRT, ssRT = self.get_io_copies()
        sim_many_dpm(self.rProb, self.rProbSS, dvg, goRT, ssRT, xtb, drift, ssdrift, bound, gbase, gOnset, ssOnset, dx, self.si, self.dt)
        if analyze:
            return self.analyze(goRT, ssRT)
        elif get_rts:
            return [goRT, ssRT]
        return pandaify_results(goRT, ssRT, ssd=self.ssd, bootstrap=False, clmap=self.clmap, tb=self.tb)


    def _simulate_traces(self, params):
        xtb, drift, ssdrift, bound, gbase, gOnset, ssOnset, dx = self.params_to_array(params, preprocess=True)
        dvg, goRT, ssRT = self.get_io_copies()
        dvs = self.dvs.copy()
        sim_many_dpm_traces(self.rProb, self.rProbSS, dvg, dvs, goRT, ssRT, xtb, drift, ssdrift, bound, gbase, gOnset, ssOnset, dx, self.si, self.dt)
        return [dvg, dvs, goRT, ssRT]


    def _simulate_ddm_traces(self, params):
        xtb, vProb, vsProb, bound, gbase, gOnset, ssOnset, dx = self.params_to_array(params, preprocess=True)
        gbase = .5*bound
        dvg, rts, _ = self.get_io_copies()
        choices = np.zeros(rts.shape)
        sim_many_ddm_traces(self.rProb, dvg, rts, choices, vProb, bound, gbase, gOnset, dx, self.dt)
        return [dvg, rts, choices]


    def analyze(self, rts, ssrts):
        """ get rt and accuracy of go and stop process for simulated
        conditions generated from simulate_dpm
        """
        nl, nssd, nssPer = ssrts.shape
        nss = nssd * nssPer
        erts = rts[:, :nss].reshape(ssrts.shape)
        gacc = np.mean(ufunc_where(rts < self.tb, 1, 0), axis=1)
        sacc = np.mean(ufunc_where(ssrts <= erts, 1, 0), axis=2)
        # rts[rts >= self.tb] = 1000.
        # erts[erts >= self.tb] = 1000.
        # ssrts[ssrts >= self.tb] =  1000.
        cq = self.RTQ(zip(rts, [self.tb] * nl))
        eq = self.RTQ(zip(erts, ssrts))
        # erts[ssrts <= erts] = 1000.
        # eq = self.RTQ(zip(erts, [self.tb] * nl))
        return hs([hs([i[ii] for i in [gacc, sacc, cq, eq]]) for ii in range(nl)])


    def params_to_array(self, params, preprocess=False):
        if type(params)==dict:
            params = self.pdict_to_array(params)
        pmat = self.pmatrix_vals.copy()
        pvary_vals = slice_theta_array(params, self.pvary_ix, self.nvary)
        pmat.loc[:, self.pvary] = pvary_vals.T
        theta_array = pmat.T.values
        if preprocess:
            return self.preproc_params(theta_array)
        return theta_array


    def pdict_to_array(self, pdict):
        plist = []
        init_ix_arrs = [np.ones(n) for n in self.nvary]
        for i, p in enumerate(self.pvary):
            plist.append(pdict[p] * init_ix_arrs[i])
        parray = np.hstack(plist)
        return parray

    def lmparams_to_array(self, lmparams):
        plist = []
        for i, pname in enumerate(self.pvary):
            p_array = np.array([p.value for p in lmparams if p.vary and pname in p.name])
            plist.append(p_array)
        parray = np.hstack(plist)
        return parray

    def preproc_params(self, theta_array):
        """
        Estimating parameters of the diffusion model: Approaches to dealing
        with contaminant reaction times and parameter variability
        Ratcliff & Tuerlinckx, 2002
        """
        a, si, sso, ssv, tr, v, xb, z = theta_array
        xtb = np.cosh(xb[:,None] * self.xtime)
        ssd = sso[:, None] + self.ssd

        # scale drift-rates
        dx = si * np.sqrt(self.dt)
        vProb = .5 * (1 + (v * np.sqrt(self.dt))/si)
        vsProb = .5 * (1 + (ssv * np.sqrt(self.dt))/si)

        # above equations give same as:
        #     s2 = si**2
        #     dx = np.sqrt(s2 * self.dt)
        #     vProb = 0.5 * (1 + v * dx / s2)
        gbase = a * z
        gOnset = get_onset_index(tr, self.dt)
        ssOnset = np.round(ssd / self.dt, 1).astype(int)
        gOnset = np.round(tr / self.dt, 1).astype(int)
        self.vProb = vProb
        self.vsProb = vsProb
        self.drift = v
        self.ssdrift = ssv
        self.gOnset = gOnset
        self.ssOnset = ssOnset
        self.dx = dx
        self.xtb = xtb
        self.bound = a
        self.gbase = gbase
        self.si = si

        return [xtb] + [v, ssv, a, gbase, gOnset, ssOnset, dx]


    def format_cond_params(self, lmParamsNames=None):
        self.ntime = np.int(np.floor(self.tb / self.dt))
        self.xtime = np.cumsum([self.dt] * self.ntime)
        self.allparams = ['a', 'si', 'sso', 'ssv', 'tr', 'v', 'xb', 'z']
        pcmap = self.fitparams['pcmap']
        apriori = pd.Series({'z': 0., 'sso': 0., 'xb': 0., 'si': self.si})
        # number of cells in condition matrix (df index)
        self.modelparams = [p for p in self.allparams if p in self.theta.keys()]
        if 'all' in list(self.depends_on) or self.fitparams.nlevels==1:
            self.pvary = np.array(self.modelparams)
            self.nvary = np.ones(len(self.pvary)).astype(np.int64)
            self.lmParamsNames = self.pvary
        else:
            self.pvary = np.array([p for p in self.allparams if p in list(self.depends_on)])
            self.nvary = np.array([len(pcmap[p]) for p in self.pvary])
            if lmParamsNames is None:
                pclist = listvalues(self.fitparams['pcmap'])
                self.lmParamsNames = np.sort(np.hstack(pclist))
        ixx = np.append(0, self.nvary.cumsum())
        self.pvary_start_ix = [(ixx[i], ixx[i+1]) for i in range(self.pvary.size)]
        self.pflat = [p for p in self.modelparams if not p in self.pvary]
        remove_def = list(set(apriori.keys()).intersection(self.pflat))
        self.apriori = apriori.drop(remove_def)
        self.make_params_matrix()
        self.set_pconstant_values_matrix(self.theta)


    def make_io_vectors(self):

        self.rProb = randsample((self.nlevels, self.ntrials, self.ntime))
        dvg = np.zeros_like(self.rProb)
        self.goRT = np.zeros((self.nlevels, self.ntrials))
        if 'ssd_info' in self.fitparams.keys():
            self.ssd, nssd, nss, nss_per, ssd_ix = self.fitparams.ssd_info
            self.rProbSS = randsample((self.nlevels, nssd, nss_per, self.ntime))
            self.rProbSS3d = randsample((self.nlevels, nssd * nss_per, self.ntime))
            self.ssRT = np.zeros((self.nlevels, nssd, nss_per))
            self.dvs = np.zeros_like(self.rProbSS)
            self.vectors = [dvg, self.goRT, self.ssRT]
            ssdSteps = get_onset_index(self.ssd, self.dt)
            self.ssdTrials = np.sort(np.tile(ssdSteps, nss_per))
            # self.ssRT2d = np.zeros((self.nlevels, nssd * nss_per))
        else:
            self.vectors = [dvg, self.goRT]

    def get_io_copies(self):
        return [v.copy() for v in self.vectors]


    def complete_allparams_pdict(self, pdict):
        theta = pd.Series(pdict)
        for p in self.allparams:
            if p not in self.modelparams:
                theta[p] = self.apriori[p]
        # order params like allparams
        theta = theta[self.allparams]
        return theta


    def set_pconstant_values_matrix(self, pdict):
        theta = self.complete_allparams_pdict(pdict)
        pmatrix_ix = self.pmatrix.copy()
        pmatrix_vals = self.pmatrix.copy()
        for p in self.allparams:
            if p in self.pvary:
                continue
            pmatrix_vals.loc[:, p] = pmatrix_ix.loc[:, p] + theta[p]
        self.pmatrix_vals = pmatrix_vals


    def pvary_broadcast_arrays(self, pmatrix):
        index = pmatrix.index.values
        c_levels = [lvl for lvl in listvalues(self.clmap)]
        level_data = list(product(*c_levels))
        condsdf = pd.DataFrame(level_data, columns=self.conds, index=index)

        # concat (along axis 1) conditional level names and depends_on param columns
        pmatrix = pd.concat([condsdf, pmatrix], axis=1)
        for pvary in self.pvary:
            pv_cond = self.depends_on[pvary]
            if isinstance(pv_cond, list):
                p_levels = self.fitparams['pcmap'][pvary]
                pvary_ix = np.arange(len(p_levels))
                pmatrix[pvary] = pvary_ix
            else:
                levels = self.clmap[pv_cond]
                for i, level_name in enumerate(levels):
                    pmatrix.ix[pmatrix[pv_cond]==level_name, pvary]=i
        pmatrix = pmatrix.iloc[:, self.nconds:].apply(pd.to_numeric)
        return pmatrix


    def make_params_matrix(self):
        cols = np.array(self.allparams)
        index = np.arange(self.nlevels)
        zerodata = np.zeros((index.size, cols.size)).astype(np.int)
        pmatrix = pd.DataFrame(zerodata, columns=cols, index=index)
        if self.nlevels>1:
            pmatrix = self.pvary_broadcast_arrays(pmatrix)
        self.pvary_ix = pmatrix[self.pvary].T.values
        self.n_vals = np.array([pmatrix[p].unique().size for p in cols])
        self.index_arrays = pmatrix.T.values
        self.pmatrix = pmatrix
